from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Object
from spinta.backends.postgresql.components import PostgreSQL


@commands.wipe.register(Context, Object, PostgreSQL)
def wipe(context: Context, dtype: Object, backend: PostgreSQL):
    for prop in dtype.properties.values():
        wipe(context, prop.dtype, backend)



from spinta.utils.schema import NA
from spinta.utils.data import take
from spinta.testing.utils import get_error_codes


def path(model: str):
    parts = model.split('/')
    if len(parts) > 3 and (parts[0], parts[2]) == ('backends', 'dtypes'):
        parts = parts[3:]
        parts = [
            a if a == 'array' else b
            for a, b in zip([None] + parts, parts)
            if b != 'array'
        ]
        return '.'.join(parts)


def nest(model: str, data: dict):
    parts = model.split('/')
    if len(parts) > 3 and (parts[0], parts[2]) == ('backends', 'dtypes'):
        parts = parts[3:]
        if parts[-1] in data:
            d = data = data.copy()
            value = d.pop(parts[-1])
            for k in parts[:-1]:
                if k == 'array':
                    v = []
                else:
                    v = {}
                if isinstance(d, list):
                    d.append(v)
                else:
                    d[k] = v
                d = v
            if isinstance(d, list):
                d.append(value)
            else:
                d[parts[-1]] = value
    return data


def flat(model, data):
    parts = model.split('/')
    if len(parts) > 3 and (parts[0], parts[2]) == ('backends', 'dtypes'):
        parts = parts[3:]
        if parts[0] in data:
            v = data = data.copy()
            for k in parts:
                if isinstance(v, list):
                    if len(v) > 0:
                        v = v[0]
                    else:
                        v = NA
                        break
                else:
                    if k in v:
                        v = v[k]
                    else:
                        v = NA
                        break
            data.pop(parts[0])
            if v is not NA:
                data[parts[-1]] = v
    return data


def post(app, model: str, value: str, *, status: int = 201):
    name = model.split('/')[-1]
    data = nest(model, {name: value})
    resp = app.post(f'/{model}', json=data)
    assert resp.status_code == status, resp.json()
    data = resp.json()
    data = flat(model, data)
    pk = data['_id']
    rev = data['_revision']
    val = data[name]
    assert data == {
        '_type': model,
        '_id': pk,
        '_revision': rev,
        name: val,
    }
    return pk, rev, val


def upsert(app, model: str, where: str, value: str, *, status: int):
    name = model.split('/')[-1]
    data = nest(model, {
        '_op': 'upsert',
        '_where': f'%s={where!r}' % path(model),
        name: value,
    })
    resp = app.post(f'/{model}', json=data)
    assert resp.status_code == status, resp.json()
    data = resp.json()
    data = flat(model, data)
    pk = data['_id']
    rev = data['_revision']
    val = data[name]
    assert data == {
        '_type': model,
        '_id': pk,
        '_revision': rev,
        name: val,
    }
    return pk, rev, val


def put(app, model: str, pk: str, rev: str, value: str = NA):
    name = model.split('/')[-1]
    data = nest(model, take(all, {
        '_revision': rev,
        name: value,
    }))
    resp = app.put(f'/{model}/{pk}', json=data)
    data = resp.json()
    assert resp.status_code == 200, data
    data = flat(model, data)
    pk = data['_id']
    rev = data['_revision']
    val = take(name, data)
    assert data == take(all, {
        '_type': model,
        '_id': pk,
        '_revision': rev,
        name: val,
    })
    return rev, val


def patch(app, model: str, pk: str, rev: str, value: str = NA):
    name = model.split('/')[-1]
    data = nest(model, take(all, {
        '_revision': rev,
        name: value,
    }))
    resp = app.patch(f'/{model}/{pk}', json=data)
    data = resp.json()
    assert resp.status_code == 200, data
    data = flat(model, data)
    pk = data['_id']
    rev = data['_revision']
    val = take(name, data)
    assert data == take(all, {
        '_type': model,
        '_id': pk,
        '_revision': rev,
        name: val,
    })
    return rev, val


def delete(app, model: str, pk: str, rev: str):
    data = nest(model, take(all, {
        '_revision': rev,
    }))
    # XXX: DELETE method should not include a request body.
    resp = app.request('DELETE', f'/{model}/{pk}', json=data)
    assert resp.content == b''
    assert resp.status_code == 204


def get(app, model, pk, rev, status=200):
    name = model.split('/')[-1]
    resp = app.get(f'{model}/{pk}?expand()')
    data = resp.json()
    assert resp.status_code == status, data
    if status == 200:
        data = flat(model, data)
        val = take(name, data)
        assert data == take(all, {
            '_type': model,
            '_id': pk,
            '_revision': rev,
            name: val,
        })
        return val
    else:
        return get_error_codes(data)


def search(app, model, pk, rev, val=NA, by=None):
    name = model.split('/')[-1]
    place = path(model)
    if val is None:
        val = 'null'
    elif isinstance(val, str):
        val = f'"{val}"'
    if by is None:
        by = f'{place}={val}'
    elif by == '_id':
        by = f'_id="{pk}"'
    else:
        by = f'{by}={val}'
    resp = app.get(f'{model}?expand()&select(_type,_id,_revision,{place})&{by}')
    data = resp.json()
    assert resp.status_code == 200, data
    assert '_data' in data, data
    val = []
    for d in data['_data']:
        d = flat(model, d)
        v = take(name, d)
        assert d == take(all, {
            '_type': model,
            '_id': pk,
            '_revision': rev,
            name: v,
        })
        val.append(v)
    return val



import base64
import json
from typing import Any, List

from spinta.components import Page, PageBy, Model, Property, UrlParams
from spinta.core.ufuncs import Env, Negative, Bind, Expr
from spinta.datasets.components import ExternalBackend
from spinta.exceptions import FieldNotInResource


class QueryPage:
    page_: Page
    size: int
    select: list
    sort: list

    def __init__(self):
        self.size = 0
        self.select = []
        self.sort = []
        self.page_ = Page()


class QueryParams:
    prioritize_uri: bool = False
    lang_priority: List[str] = None
    push: bool = False


class BaseQueryBuilder(Env):
    page: QueryPage
    expand: List[Property] = None
    query_params: QueryParams = None

    def init_query_params(self, params: QueryParams):
        if params is None:
            params = QueryParams()
        self.query_params = params


class LoadBuilder(Env):
    model: Model

    def resolve(self, expr: Any):
        if not isinstance(expr, Expr):
            # Expression is already resolved, return resolved value.
            return expr

        if expr.name in self._resolvers:
            ufunc = self._resolvers[expr.name]

        else:
            args, kwargs = expr.resolve(self)
            return self.default_resolver(expr, *args, **kwargs)

        if ufunc.autoargs:
            # Resolve arguments automatically.
            args, kwargs = expr.resolve(self)
            try:
                return ufunc(self, *args, **kwargs)
            except NotImplementedError:
                return self.default_resolver(expr, *args, **kwargs)

        else:
            # Resolve arguments manually.
            try:
                return ufunc(self, expr)
            except NotImplementedError:
                pass

    def load_page(self):
        page = Page()
        page_given = False
        if self.model.external and self.model.external.prepare:
            resolved = self.resolve(self.model.external.prepare)
            if not isinstance(resolved, list):
                resolved = [resolved]
            for item in resolved:
                if isinstance(item, Page):
                    page = item
                    page_given = True
                    break
        if not page_given:
            args = ['_id']
            if self.model.given.pkeys:
                if isinstance(self.model.given.pkeys, list):
                    args = self.model.given.pkeys
                else:
                    args = [self.model.given.pkeys]
                if '_id' in args:
                    args.remove('_id')
            for arg in args:
                key = arg
                if arg in self.model.properties:
                    prop = self.model.properties[arg]
                    page.by.update({
                        key: PageBy(prop)
                    })
                else:
                    raise FieldNotInResource(self.model, property=arg)

        # Disable page if given properties are not possible to access
        for page_by in page.by.values():
            if not isinstance(page_by.prop.dtype, get_allowed_page_property_types()):
                page.is_enabled = False
                break
        self.model.page = page


def get_allowed_page_property_types():
    from spinta.types.datatype import Integer, Number, String, Date, Time, DateTime, PrimaryKey
    return Integer, Number, String, Date, DateTime, Time, PrimaryKey


def get_page_values(env: BaseQueryBuilder, row: dict):
    if not env.page.page_.filter_only:
        if isinstance(env.model.backend, ExternalBackend):
            return [row[item.prop.external.name] for item in env.page.page_.by.values()]
        else:
            return [row[item.prop.name] for item in env.page.page_.by.values()]


def merge_with_page_selected_list(select_list: list, page: QueryPage):
    merged_selected = select_list or []
    if page.page_.is_enabled and page.select is not None:
        for select in page.select:
            if select not in merged_selected:
                merged_selected.append(select)
    return merged_selected


def merge_with_page_sort(sort: list, page: QueryPage):
    merged_sort = sort or []
    if page.page_.is_enabled and page.sort:
        merged_sort = page.sort
    return merged_sort


def merge_with_page_limit(limit: int, page: QueryPage):
    if page and page.size and page.page_.is_enabled:
        if limit and limit < page.size:
            return limit
        return page.size
    return limit


def update_query_with_url_params(query_params: QueryParams, url_params: UrlParams):
    query_params.prioritize_uri = url_params.fmt.prioritize_uri
    query_params.lang_priority = url_params.accept_langs



import importlib
import pathlib
from typing import Type
from typing import TypeVar

from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.core.config import configure_rc

from spinta.core.config import read_config
from spinta.utils.imports import importstr


ContextType = TypeVar('ContextType', bound=Context)


def create_context(
    name='spinta',
    rc=None,
    context: ContextType = None,
    args=None,
    envfile=None,
) -> ContextType:
    if rc is None:
        rc = read_config(args, envfile)

    load_commands(rc.get('commands', 'modules', cast=list))

    if context is None:
        Context_: Type[Context] = rc.get('components', 'core', 'context', cast=importstr, required=True)
        context = Context_(name)

    context.set('rc', rc)

    Config = rc.get('components', 'core', 'config', cast=importstr, required=True)
    context.set('config', Config())

    Store = rc.get('components', 'core', 'store', cast=importstr, required=True)
    context.set('store', Store())

    return context


def load_commands(modules):
    for module_path in modules:
        module = importlib.import_module(module_path)
        path = pathlib.Path(module.__file__).resolve()
        if path.name != '__init__.py':
            continue
        path = path.parent
        base = path.parents[module_path.count('.')]
        for path in path.glob('**/*.py'):
            if path.name == '__init__.py':
                module_path = path.parent.relative_to(base)
            else:
                module_path = path.relative_to(base).with_suffix('')
            module_path = '.'.join(module_path.parts)
            module = importlib.import_module(module_path)


def configure_context(context: Context, *args, **kwargs) -> Context:
    rc: RawConfig = context.get('rc')
    context = context.fork('configure')
    context.set('rc', configure_rc(rc, *args, **kwargs))
    return context



import pathlib
import json
from copy import deepcopy

import requests
import xmltodict
from typing import List, Dict, Union, TypedDict, Any, Tuple, Callable

from spinta.manifests.dict.components import DictFormat
from spinta.manifests.helpers import TypeDetector
from spinta.utils.naming import Deduplicator, to_model_name, to_property_name


def read_schema(manifest_type: DictFormat, path: str):
    if path.startswith(('http://', 'https://')):
        value = requests.get(path).text
    else:
        with pathlib.Path(path).open(encoding='utf-8-sig') as f:
            value = f.read()

    converted = {}
    mapping_meta = _MappingMeta(
        is_blank_node=False,
        blank_node_name="model1",
        blank_node_source=".",
        seperator="",
        recursive_descent="",
        model_source_prefix="",
        namespace_seperator=":",
        remove_array_suffix=False,
        check_namespace=False,
        namespace_prefixes={},
        detect_model=is_model
    )
    if manifest_type == DictFormat.JSON:
        converted = json.loads(value)
        mapping_meta["seperator"] = "."
        mapping_meta["recursive_descent"] = "."

    elif manifest_type in (DictFormat.XML, DictFormat.HTML):
        converted = xmltodict.parse(value)
        mapping_meta["seperator"] = "/"
        mapping_meta["recursive_descent"] = "/.."
        mapping_meta["model_source_prefix"] = "/"
        mapping_meta["remove_array_suffix"] = True
        mapping_meta["check_namespace"] = True
        mapping_meta["namespace_prefixes"] = {
            "xmlns": ["xmlns", "@xmlns"]
        }

    namespaces = extract_namespaces(converted, mapping_meta)
    prefixes = {}
    for i, (key, value) in enumerate(namespaces):
        prefixes[key] = {
            'type': 'prefix',
            'name': key,
            'uri': value,
            'eid': i
        }
    dataset_structure: _MappedDataset = {
        "dataset": "dataset",
        "resource": "resource",
        "models": {}
    }
    mapping_meta["is_blank_node"] = is_blank_node(converted)
    create_type_detectors(dataset_structure, converted, mapping_meta)

    yield None, {
        'type': 'dataset',
        'name': dataset_structure["dataset"],
        'prefixes': prefixes,
        'resources': {
            dataset_structure["resource"]: {
                'type': manifest_type.value,
                'external': path,
            },
        },
    }

    dedup_model = Deduplicator('{}')
    blank_model = ""
    mapped_models: Dict[Tuple, str] = {}

    for model in dataset_structure["models"].values():
        for model_source, m in model.items():
            new_model = to_model_name(_name_without_namespace(m["name"], mapping_meta, prefixes))
            new_model = dedup_model(new_model)
            mapped_models[(m['name'], model_source)] = new_model

    for model in dataset_structure["models"].values():
        for model_source, m in model.items():
            if model_source == mapping_meta["blank_node_source"]:
                blank_model = mapped_models[(m["name"], model_source)]
            dedup_prop = Deduplicator('_{}')
            converted_props = {}
            for prop in m["properties"].values():
                new_prop = to_property_name(_name_without_namespace(prop["name"], mapping_meta, prefixes))
                new_prop = dedup_prop(new_prop)
                extra = {}
                type_detector = prop["type_detector"]
                prop_type = type_detector.get_type()
                if prop_type == "ref":
                    if prop['name'] in dataset_structure["models"].keys():
                        model_name = _name_without_namespace(mapped_models[prop['name'], prop['extra']], mapping_meta, prefixes)
                        ref_model = f'{dataset_structure["dataset"]}/{model_name}'
                    else:
                        ref_model = f'{dataset_structure["dataset"]}/{blank_model}'
                    extra = {
                        'model': ref_model
                    }
                converted_props[new_prop] = {
                    'type': prop_type,
                    'external': {
                        'name': prop["source"]
                    },
                    'description': '',
                    'required': type_detector.required,
                    'unique': type_detector.unique,
                    **extra
                }
                if type_detector.array:
                    copied = deepcopy(converted_props[new_prop])
                    converted_props[new_prop]['type'] = 'array'
                    converted_props[new_prop]['explicitly_given'] = False
                    copied['given_name'] = f'{new_prop}[]'
                    converted_props[new_prop]['items'] = copied

            fixed_external_source = m["source"]
            if mapping_meta["remove_array_suffix"]:
                fixed_external_source = fixed_external_source.replace("[]", "")
            if not fixed_external_source.startswith("."):
                fixed_external_source = f'{mapping_meta["model_source_prefix"]}{fixed_external_source}'
            yield None, {
                'type': 'model',
                'name': f'{dataset_structure["dataset"]}/{mapped_models[(m["name"], model_source)]}',
                'external': {
                    'dataset': dataset_structure["dataset"],
                    'resource': dataset_structure["resource"],
                    'name': fixed_external_source,
                },
                'description': '',
                'properties': converted_props
            }


class _MappedProperties(TypedDict):
    name: str
    source: str
    extra: str
    type_detector: TypeDetector


class _MappedModels(TypedDict):
    name: str
    source: str
    properties: Dict[str, _MappedProperties]


class _MappedDataset(TypedDict):
    dataset: str
    resource: str
    models: Dict[str, Dict[str, _MappedModels]]


class _MappingMeta(TypedDict):
    is_blank_node: bool
    blank_node_name: str
    blank_node_source: str
    seperator: str
    recursive_descent: str
    remove_array_suffix: bool
    model_source_prefix: str
    check_namespace: bool
    namespace_prefixes: dict
    namespace_seperator: str
    detect_model: Callable


class _MappingScope(TypedDict):
    parent_scope: str
    model_scope: str
    model_name: str
    property_name: str


def _name_without_namespace(name: str, mapping_meta: _MappingMeta, prefixes: dict):
    if mapping_meta["namespace_seperator"] in name:
        for prefix in prefixes.keys():
            namespace = f'{prefix}{mapping_meta["namespace_seperator"]}'
            if namespace in name:
                return name.replace(namespace, '')
    return name


def is_model(data):
    if isinstance(data, list) and is_list_of_dicts(data):
        return True
    return False


def _find_parent_key(data, string):
    for key, values in data.items():
        for value in values:
            if string.startswith(value):
                return key
    return None


def extract_namespaces(data: Any, mapping_meta: _MappingMeta):
    if mapping_meta["check_namespace"]:
        if isinstance(data, dict):
            keys_to_remove = {}
            for key, value in data.items():
                parent = _find_parent_key(mapping_meta["namespace_prefixes"], key)
                if parent is not None:
                    if key not in keys_to_remove.keys():
                        return_key = key.split(mapping_meta["namespace_seperator"])
                        if len(return_key) == 1:
                            return_key = [parent]
                        keys_to_remove[key] = (return_key[-1], value)
            for key, value in keys_to_remove.items():
                del data[key]
                yield value
            for value in data.values():
                yield from extract_namespaces(value, mapping_meta)
        elif isinstance(data, list):
            for item in data:
                yield from extract_namespaces(item, mapping_meta)


def nested_prop_names(new_values: list, values: dict, root: str, seperator: str):
    for key, value in values.items():
        if isinstance(value, dict):
            nested_prop_names(new_values, value, f'{root}{seperator}{key}', seperator)
        elif isinstance(value, list):
            if not is_list_of_dicts(value):
                new_values.append(f'{root}{seperator}{key}')
        else:
            new_values.append(f'{root}{seperator}{key}')


def check_missing_prop_required(dataset: _MappedDataset, values: dict, mapping_scope: _MappingScope, mapping_meta: _MappingMeta):
    if mapping_scope["model_scope"] == "":
        model_name = mapping_scope["model_name"] if mapping_scope["model_name"] != "" else list(dataset["models"].keys())[0]
        model_source = _create_name_with_prefix(
            mapping_scope["parent_scope"],
            mapping_meta["seperator"],
            model_name
        )
        if mapping_scope["model_name"] == "" and mapping_meta["is_blank_node"]:
            model_name = mapping_meta["blank_node_name"]
            model_source = mapping_meta["blank_node_source"]
        key_values = []
        for k, v in values.items():
            new_val = _create_name_with_prefix(
                mapping_scope["model_scope"],
                mapping_meta["seperator"],
                k
            )
            if isinstance(v, dict):
                nested_prop_names(key_values, v, new_val, mapping_meta['seperator'])
            elif isinstance(v, list):
                if not is_list_of_dicts(v):
                    key_values.append(new_val)
            else:
                key_values.append(new_val)
        if key_values:
            filtered_models = dataset["models"][model_name][model_source]["properties"].keys()
            subtracted = set(filtered_models) - set(key_values)
            for prop in subtracted:
                type_detector = dataset["models"][model_name][model_source]["properties"][prop]["type_detector"]
                if type_detector.required and type_detector.type != 'ref':
                    type_detector.required = False


def run_type_detectors(dataset: _MappedDataset, values: dict, mapping_scope: _MappingScope, mapping_meta: _MappingMeta):
    if isinstance(values, list):
        for item in values:
            run_type_detectors(dataset, item, mapping_scope, mapping_meta)
    elif isinstance(values, dict):
        check_missing_prop_required(dataset, values, mapping_scope, mapping_meta)
        for key, value in values.items():
            new_mapping_scope = _MappingScope(
                parent_scope=mapping_scope["parent_scope"],
                model_name=mapping_scope["model_name"],
                model_scope=mapping_scope["model_scope"],
                property_name=key
            )
            if mapping_meta["detect_model"](value):
                new_mapping_scope["model_name"] = key
                parent_scope = _create_name_with_prefix(
                    new_mapping_scope["parent_scope"],
                    mapping_meta["seperator"],
                    "" if mapping_scope["model_name"] == "" else f'{mapping_scope["model_name"]}[]'
                )
                parent_scope = _create_name_with_prefix(
                    parent_scope,
                    mapping_meta["seperator"],
                    new_mapping_scope["model_scope"]
                )
                new_mapping_scope["model_scope"] = ""
                new_mapping_scope["parent_scope"] = parent_scope
                run_type_detectors(dataset, value, new_mapping_scope, mapping_meta)
            else:
                if isinstance(value, dict):
                    new_mapping_scope["model_scope"] = _create_name_with_prefix(
                        new_mapping_scope["model_scope"],
                        mapping_meta["seperator"],
                        key
                    )
                    run_type_detectors(dataset, value, new_mapping_scope, mapping_meta)
                else:
                    _detect_type(dataset, new_mapping_scope, mapping_meta, value)


def _detect_type(dataset: _MappedDataset, mapping_scope: _MappingScope, mapping_meta: _MappingMeta, value: Any):
    prop_name = _create_name_with_prefix(
        mapping_scope["model_scope"],
        mapping_meta["seperator"],
        mapping_scope["property_name"]
    )
    model_name = mapping_scope["model_name"]
    model_source = _create_name_with_prefix(
        mapping_scope["parent_scope"],
        mapping_meta["seperator"],
        mapping_scope["model_name"]
    )
    if model_name == "":
        model_name = mapping_meta["blank_node_name"]
        model_source = mapping_meta["blank_node_source"]
    dataset["models"][model_name][model_source]["properties"][prop_name]["type_detector"].detect(value)


def is_list_of_dicts(lst: List) -> bool:
    for item in lst:
        if not isinstance(item, dict):
            return False
    return True


def setup_model_type_detectors(dataset: _MappedDataset, values: Union[dict, list], mapping_scope: _MappingScope,
                               mapping_meta: _MappingMeta):
    if isinstance(values, list):
        for item in values:
            setup_model_type_detectors(dataset, item, mapping_scope, mapping_meta)
    elif isinstance(values, dict):
        for key, value in values.items():
            new_mapping_scope = _MappingScope(
                parent_scope=mapping_scope["parent_scope"],
                model_name=mapping_scope["model_name"],
                model_scope=mapping_scope["model_scope"],
                property_name=key
            )
            if mapping_meta["detect_model"](value):
                new_mapping_scope["model_name"] = key
                parent_scope = _create_name_with_prefix(
                    new_mapping_scope["parent_scope"],
                    mapping_meta["seperator"],
                    "" if mapping_scope["model_name"] == "" else f'{mapping_scope["model_name"]}[]'
                )
                parent_scope = _create_name_with_prefix(
                    parent_scope,
                    mapping_meta["seperator"],
                    new_mapping_scope["model_scope"]
                )
                new_mapping_scope["model_scope"] = ""
                new_mapping_scope["parent_scope"] = parent_scope

                setup_model_type_detectors(dataset, value, new_mapping_scope, mapping_meta)

                old_mapping_scope = new_mapping_scope.copy()
                old_mapping_scope["property_name"] = mapping_scope["property_name"]

                if old_mapping_scope["property_name"] == "" and mapping_meta["is_blank_node"]:
                    old_mapping_scope["property_name"] = "parent"
                    set_type_detector(dataset, old_mapping_scope, mapping_meta, is_ref=True)
                elif old_mapping_scope["property_name"] != "":
                    set_type_detector(dataset, old_mapping_scope, mapping_meta, is_ref=True)
            else:
                if isinstance(value, list):
                    set_type_detector(dataset, new_mapping_scope, mapping_meta, is_array=True)
                elif isinstance(value, dict):
                    new_mapping_scope["model_scope"] = _create_name_with_prefix(
                        new_mapping_scope["model_scope"],
                        mapping_meta["seperator"],
                        key
                    )
                    new_mapping_scope["property_name"] = mapping_scope["property_name"]
                    setup_model_type_detectors(dataset, value, new_mapping_scope, mapping_meta)
                else:
                    set_type_detector(dataset, new_mapping_scope, mapping_meta)


def _create_name_with_prefix(prefix: str, seperator: str, value: str) -> str:
    if value == "":
        return prefix
    return value if prefix == "" else f'{prefix}{seperator}{value}'


def create_type_detectors(dataset: _MappedDataset, values: Any, mapping_meta: _MappingMeta):
    mapping_scope = _MappingScope(
        parent_scope="",
        model_scope="",
        model_name="",
        property_name=""
    )
    setup_model_type_detectors(dataset, values, mapping_scope, mapping_meta)
    run_type_detectors(dataset, values, mapping_scope, mapping_meta)


def is_blank_node(values: Union[list, dict]) -> bool:
    if isinstance(values, list):
        return True
    if isinstance(values, dict):
        val = values
        first_value = val[list(val.keys())[0]]
        if len(val.keys()) == 1 and isinstance(first_value, dict):
            while len(val.keys()) == 1 and isinstance(first_value, dict):
                val = val[list(val.keys())[0]]
                first_value = val[list(val.keys())[0]]
                if len(val.keys()) != 1:
                    return True
                if isinstance(first_value, list):
                    return False
            return True
        else:
            for key, value in values.items():
                if not isinstance(value, list):
                    return True
    return False


def set_type_detector(dataset: _MappedDataset, mapping_scope: _MappingScope, mapping_meta: _MappingMeta,
                      is_ref: bool = False, is_array: bool = False):
    if mapping_scope["property_name"] != '':
        model_name = mapping_scope["model_name"]
        model_source = _create_name_with_prefix(
            mapping_scope["parent_scope"],
            mapping_meta["seperator"],
            model_name
        )
        prop_name = _create_name_with_prefix(
            mapping_scope["model_scope"],
            mapping_meta["seperator"],
            mapping_scope["property_name"]
        )
        prop_source = prop_name
        if model_name == "" and mapping_meta["is_blank_node"]:
            model_name = mapping_meta["blank_node_name"]
            model_source = mapping_meta["blank_node_source"]
        extra = ""
        if is_ref:
            extra = ""
            split = mapping_scope["parent_scope"].split(mapping_meta["seperator"])
            count = 0
            for i in reversed(split):
                if i.endswith("[]"):
                    half = mapping_scope["parent_scope"].split(i)
                    extra = i[:-2] if half[0] == '' else f'{half[0]}{i[:-2]}'
                    break
                elif i != '':
                    count += 1
            prop_source = f"..{mapping_meta['recursive_descent'] * count}"

        if model_name in dataset["models"].keys():
            if model_source in dataset["models"][model_name].keys():
                if prop_name not in dataset["models"][model_name][model_source]["properties"].keys():
                    dataset["models"][model_name][model_source]["properties"][prop_name] = {
                        "name": prop_name,
                        "source": prop_source,
                        "type_detector": TypeDetector(),
                        'extra': extra
                    }
            else:
                dataset["models"][model_name][model_source] = {
                    "name": model_name,
                    "source": model_source,
                    "properties": {
                        prop_name: {
                            "name": prop_name,
                            "source": prop_source,
                            "type_detector": TypeDetector(),
                            'extra': extra
                        }
                    }
                }
        else:
            dataset["models"][model_name] = {
                model_source: {
                    "name": model_name,
                    "source": model_source,
                    "properties": {
                        prop_name: {
                            "name": prop_name,
                            "source": prop_source,
                            "type_detector": TypeDetector(),
                            'extra': extra
                        }
                    }
                }
            }

        type_detector = dataset["models"][model_name][model_source]["properties"][prop_name]["type_detector"]
        if is_array:
            type_detector.array = True
            type_detector.unique = False
            type_detector.required = False
        if is_ref:
            type_detector.type = 'ref'
            type_detector.unique = False
            type_detector.required = False



from typing import Any
from typing import Union

from spinta.components import Context
from spinta.components import Model
from spinta.components import Property
from spinta.core.ufuncs import Bind, asttoexpr
from spinta.core.ufuncs import Expr
from spinta.exceptions import PropertyNotFound
from spinta.manifests.components import Manifest
from spinta.naming.components import NameFormatter
from spinta.spyna import parse
from spinta.types.datatype import Ref
from spinta.ufuncs.components import ForeignProperty
from spinta.utils.naming import to_model_name
from spinta.utils.naming import to_property_name


def _re_format_expr(env: NameFormatter, expr: Any) -> Any:
    if isinstance(expr, Expr):
        args = [_re_format_expr(env, a) for a in expr.args]
        kwargs = {k: _re_format_expr(env, v) for k, v in expr.kwargs.items()}
        return type(expr)(expr.name, *args, **kwargs)

    if isinstance(expr, tuple):
        return tuple(_re_format_expr(env, v) for v in expr)

    if isinstance(expr, list):
        return [_re_format_expr(env, v) for v in expr]

    if isinstance(expr, dict):
        return {k: _re_format_expr(env, v) for k, v in expr}

    if isinstance(expr, Bind):
        if expr.name not in env.model.properties:
            raise PropertyNotFound(env.model, property=expr.name)
        prop = env.model.properties[expr.name]
        return Expr('bind', _format_property_place(prop))

    if isinstance(expr, ForeignProperty):
        fpr: ForeignProperty = expr
        expr = Expr(
            'getattr',
            Expr('bind', _format_property_place(fpr.left.prop)),
            Expr('bind', _format_property_place(fpr.right.prop)),
        )
        for fpr in reversed(fpr.chain[:-1]):
            expr = Expr(
                'getattr',
                expr,
                Expr('bind', _format_property_place(fpr.left.prop)),
            )
        return expr

    return expr


def _format_expr(
    context: Context,
    node: Union[Model, Property],
    expr: Expr,
) -> Expr:
    env = NameFormatter(context)
    if isinstance(node, Model):
        env.update(model=node, prop=None)
    elif isinstance(node, Property):
        env.update(model=node.model, prop=node)
    else:
        raise NotImplementedError
    expr = env.resolve(expr)
    return _re_format_expr(env, expr)


def _format_property_expr(context: Context, prop: Property) -> None:
    if prop.external and prop.external.prepare:
        prop.external.prepare = _format_expr(context, prop, prop.external.prepare)
    if prop.given.prepare:
        for item in prop.given.prepare:
            if not item['appended']:
                item['prepare'] = _format_expr(context, prop, asttoexpr(parse(item['prepare'])))


def _format_model_expr(context: Context, model: Model) -> None:
    if model.external and model.external.prepare:
        model.external.prepare = _format_expr(context, model, model.external.prepare)
    for prop in model.properties.values():
        _format_property_expr(context, prop)


def _format_model_name(name: str) -> str:
    if '/' in name:
        ns, name = name.rsplit('/', 1)
        return '/'.join([ns.lower(), to_model_name(name)])
    else:
        return to_model_name(name)


def _format_property_name(prop: Property) -> str:
    is_ref = isinstance(prop.dtype, Ref)
    return to_property_name(prop.name, is_ref)


def _format_property_given_name(prop: Property) -> str:
    is_ref = isinstance(prop.dtype, Ref)
    return to_property_name(prop.given.name, is_ref)


def _format_property_place(prop: Property) -> str:
    is_ref = isinstance(prop.dtype, Ref)
    place = prop.place.split('.')
    return '.'.join(
        [to_property_name(n) for n in place[:-1]] +
        [to_property_name(n, is_ref) for n in place[-1:]]
    )


def _format_property(prop: Property) -> Property:
    if prop.name.startswith('_'):
        return prop
    else:
        prop.name = _format_property_name(prop)
        prop.place = _format_property_place(prop)
        prop.given.name = _format_property_given_name(prop)
        return prop


def _format_model(model: Model) -> Model:
    model.name = _format_model_name(model.name)
    model.properties = {
        prop.name: prop
        for prop in map(_format_property, model.properties.values())
    }
    return model


def reformat_names(context: Context, manifest: Manifest):
    for model in manifest.models.values():
        _format_model_expr(context, model)

    manifest.objects['model'] = {
        model.name: model
        for model in map(_format_model, manifest.models.values())
    }



from spinta import commands
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.manifests.rdf.components import RdfManifest


@commands.configure.register(Context, RdfManifest)
def configure(context: Context, manifest: RdfManifest):
    rc: RawConfig = context.get('rc')
    manifest.path = rc.get('manifests', manifest.name, 'path')
    if manifest.path.endswith('.rdf'):
        manifest.format = 'application/rdf+xml'
    else:
        manifest.format = 'turtle'



import pathlib


def is_ignored(rules, base, path):
    path = pathlib.Path('/') / path.relative_to(base)
    for rule in rules:
        if rule.endswith('/'):
            rule_parts_count = rule.count('/')
            if rule.startswith('/'):
                stars = len(path.parts) - rule_parts_count
                if stars > 0 and path.match(rule + '/'.join(['*'] * stars)):
                    return True
            else:
                for i in range(len(path.parts)):
                    p = pathlib.Path(*path.parts[i:])
                    stars = len(p.parts) - rule_parts_count
                    if stars > 0 and p.match(rule + '/'.join(['*'] * stars)):
                        return True
            if rule and path.match(rule):
                return True
        else:
            if path.match(rule):
                return True
    return False



from __future__ import annotations

import uuid
from typing import List, Union, Any, Tuple

import datetime
import dataclasses
from typing import Optional

import sqlalchemy as sa

from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.dialects.postgresql import array_agg

from spinta import exceptions
from spinta.auth import authorized
from spinta.backends import get_property_base_model
from spinta.core.ufuncs import ufunc
from spinta.core.ufuncs import Bind, Negative as Negative_
from spinta.core.ufuncs import Expr
from spinta.datasets.enums import Level
from spinta.exceptions import EmptyStringSearch, PropertyNotFound, LangNotDeclared
from spinta.exceptions import UnknownMethod
from spinta.exceptions import FieldNotInResource
from spinta.components import Model, Property, Action, Page, UrlParams
from spinta.ufuncs.basequerybuilder.components import BaseQueryBuilder, QueryPage, merge_with_page_sort, \
    merge_with_page_limit, merge_with_page_selected_list, QueryParams
from spinta.utils.data import take
from spinta.types.datatype import DataType, ExternalRef, Inherit, BackRef, Time, ArrayBackRef
from spinta.types.datatype import Array
from spinta.types.datatype import File
from spinta.types.datatype import Object
from spinta.types.datatype import Ref
from spinta.types.datatype import String
from spinta.types.datatype import Integer
from spinta.types.datatype import Number
from spinta.types.datatype import DateTime
from spinta.types.datatype import Date
from spinta.types.datatype import PrimaryKey
from spinta.types.text.components import Text
from spinta.backends.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.components import BackendFeatures


class PgQueryBuilder(BaseQueryBuilder):
    backend: PostgreSQL

    def init(self, backend: PostgreSQL, table: sa.Table, params: QueryParams = None):
        result = self(
            backend=backend,
            table=table,
            # TODO: Select list must be taken from params.select.
            select=None,
            joins={},
            from_=table,
            sort=[],
            limit=None,
            offset=None,
            aggregate=False,
            page=QueryPage(),
            group_by=[],
        )
        result.init_query_params(params)
        return result

    def build(self, where):
        if self.select is None:
            self.call('select', Expr('select'))

        if self.aggregate:
            select = []
        else:
            select = [
                self.table.c['_id'],
                self.table.c['_revision'],
            ]
        merged_selected = merge_with_page_selected_list(list(self.select.values()), self.page)
        merged_sorted = merge_with_page_sort(self.sort, self.page)
        merged_limit = merge_with_page_limit(self.limit, self.page)
        for sel in merged_selected:
            if sel is not None:
                if sel.prop and sel.prop.dtype.expandable:
                    if self.expand is None or self.expand and sel.prop not in self.expand:
                        continue
                items = sel.item if isinstance(sel.item, list) else [sel.item]
                for item in items:
                    if item is not None and item not in select:
                        select.append(item)

        qry = sa.select(select)
        qry = qry.select_from(self.from_)

        if where is not None:
            qry = qry.where(where)

        if merged_sorted:
            qry = qry.order_by(*merged_sorted)

        if merged_limit is not None:
            qry = qry.limit(merged_limit)

        if self.offset is not None:
            qry = qry.offset(self.offset)

        if self.group_by:
            qry = qry.group_by(*self.group_by)
        return qry

    def default_resolver(self, expr, *args, **kwargs):
        raise UnknownMethod(expr=repr(expr(*args, **kwargs)), name=expr.name)

    def get_joined_table(self, prop: ForeignProperty) -> sa.Table:
        for fpr in prop.chain:
            if fpr.name in self.joins:
                continue

            if len(fpr.chain) > 1:
                ltable = self.joins[fpr.chain[-2].name]
            else:
                ltable = self.backend.get_table(fpr.left.model)
            lrkey = self.backend.get_column(ltable, fpr.left)

            rmodel = fpr.right.model
            rtable = self.backend.get_table(rmodel).alias()
            rpkey = self.backend.get_column(rtable, rmodel.properties['_id'])

            condition = lrkey == rpkey
            self.joins[fpr.name] = rtable
            self.from_ = self.from_.outerjoin(rtable, condition)

        return self.joins[fpr.name]

    def generate_backref_select(self, left: Property, right: Property, required_columns: List[Tuple[str, str]], aggregate: bool = False):
        select_columns = []
        group_by = []

        left_table = self.backend.get_table(left.model)
        left_keys = self.backend.get_column(left_table, left)

        join_condition = None
        right_table = self.backend.get_table(right)
        right_list_table = None
        list_join_condition = None
        right_keys = self.backend.get_column(right_table, right, override_table=False)
        right_list_keys = None
        if right.list is not None:
            right_list_table = self.backend.get_table(right, TableType.LIST)
            list_join_condition = right_list_table.c['_rid'] == right_table.c['_id']
            right_list_keys = self.backend.get_column(right_list_table, right)
        from_ = right_list_table if right_list_table is not None else right_table
        initial_join_right_keys = right_list_keys if right_list_keys is not None else right_keys

        if aggregate:
            json_columns = []
            for col, label in required_columns:
                json_columns.extend([col, right_table.c[col]])
            column = array_agg(sa.func.json_build_object(*json_columns))
            column = column.label(left.name)
            select_columns.append(column)

            if isinstance(initial_join_right_keys, list):
                for key in initial_join_right_keys:
                    group_by.append(key)
            else:
                group_by.append(initial_join_right_keys)

        else:
            for col, label in required_columns:
                column = right_table.c[col]
                column = column.label(label)
                select_columns.append(column)

        if isinstance(initial_join_right_keys, list):
            select_columns.extend(initial_join_right_keys)
        else:
            select_columns.append(initial_join_right_keys)

        if isinstance(left_keys, list) and isinstance(initial_join_right_keys, list):
            if len(left_keys) == len(initial_join_right_keys):
                for i, key in enumerate(left_keys):
                    if join_condition is None:
                        join_condition = key == initial_join_right_keys[i]
                    else:
                        join_condition = sa.and_(join_condition, key == initial_join_right_keys[i])
            else:
                raise Exception("DOESNT MATCH")
        elif type(left_keys) != type(initial_join_right_keys):
            raise Exception("TYPE DOESNT MATCH")
        else:
            join_condition = left_keys == initial_join_right_keys

        from_ = from_.outerjoin(left_table, join_condition)
        if list_join_condition is not None:
            from_ = from_.outerjoin(right_table, list_join_condition)
        stmt = sa.select(select_columns)
        stmt = stmt.select_from(from_)
        stmt = stmt.group_by(*group_by)
        stmt = stmt.subquery()
        return stmt

    def get_backref_joined_table(self, prop: ForeignProperty, selector: sa.sql.expression) -> sa.sql.expression:
        for fpr in prop.chain:
            if fpr.name in self.joins:
                continue
            if len(fpr.chain) > 1:
                ltable = self.joins[fpr.chain[-2].name]
            else:
                ltable = self.backend.get_table(fpr.left.model)
            lrkey = self.backend.get_column(ltable, fpr.left)
            rpkey = self.backend.get_column(selector, fpr.right, override_table=False)
            condition = None
            if isinstance(lrkey, list) and isinstance(rpkey, list):
                if len(lrkey) == len(rpkey):
                    for i, key in enumerate(lrkey):
                        if condition is None:
                            condition = key == rpkey[i]
                        else:
                            condition = sa.and_(condition, key == rpkey[i])
                else:
                    raise Exception("DOESNT MATCH")
            elif type(lrkey) != type(rpkey):
                raise Exception("TYPE DOESNT MATCH")
            else:
                condition = lrkey == rpkey
            self.joins[fpr.name] = selector
            self.from_ = self.from_.outerjoin(selector, condition)
        return self.joins[fpr.name]

    def get_joined_base_table(self, model: Model, prop: str):
        inherit_model = model
        base_model = get_property_base_model(inherit_model, prop)

        if not base_model:
            raise PropertyNotFound(prop)

        if base_model.name in self.joins:
            return self.joins[base_model.name]

        ltable = self.backend.get_table(inherit_model)
        lrkey = self.backend.get_column(ltable, inherit_model.properties['_id'])

        rtable = self.backend.get_table(base_model).alias()
        rpkey = self.backend.get_column(rtable, base_model.properties['_id'])

        condition = lrkey == rpkey
        self.joins[base_model.name] = rtable
        self.from_ = self.from_.outerjoin(rtable, condition)

        return self.joins[base_model.name]


def _gather_selected_properties(env: PgQueryBuilder):
    result = []
    if env.select:
        for selected in env.select.values():
            if selected and selected.prop:
                result.append(selected.prop)
    return result


class ForeignProperty:

    def __init__(
        self,
        fpr: Optional[ForeignProperty],
        left: Property,
        right: Property,
    ):
        if fpr is None:
            self.name = left.place
            self.chain = [self]
        else:
            self.name = fpr.name + '->' + left.place
            self.chain = fpr.chain + [self]

        self.left = left
        self.right = right

    def __repr__(self):
        return (
            'ForeignProperty('
            f'{self.name}->{self.right.name}:{self.right.dtype.name}'
            ')'
        )

    @property
    def place(self):
        return '.'.join(
            [fpr.left.place for fpr in self.chain] +
            [self.right.place]
        )


class InheritForeignProperty:

    def __init__(
        self,
        model: Model,
        prop_name: str,
        base_prop: Property
    ):
        self.model = model
        self.base_prop = base_prop
        self.prop_name = prop_name


class Func:
    pass


@dataclasses.dataclass
class Lower(Func):
    dtype: DataType = None


@dataclasses.dataclass
class Negative(Func):
    arg: Any


@dataclasses.dataclass
class Positive(Func):
    arg: Any


class Star:
    pass


@dataclasses.dataclass
class Recurse(Func):
    args: List[Union[DataType, Func]] = None


@dataclasses.dataclass
class ReservedProperty(Func):
    dtype: DataType
    param: str


@ufunc.resolver(PgQueryBuilder, str, name='op')
def op_(env, arg: str):
    if arg == '*':
        return Star()
    else:
        raise NotImplementedError


@ufunc.resolver(PgQueryBuilder, Bind, Bind, name='getattr')
def getattr_(env, field, attr):
    if field.name in env.model.properties:
        prop = env.model.properties[field.name]
    else:
        raise FieldNotInResource(env.model, property=field.name)
    return env.call('getattr', prop.dtype, attr)


@ufunc.resolver(PgQueryBuilder, Object, Bind, name='getattr')
def getattr_(env, dtype, attr):
    if attr.name in dtype.properties:
        return dtype.properties[attr.name].dtype
    else:
        raise FieldNotInResource(dtype, property=attr.name)


@ufunc.resolver(PgQueryBuilder, Array, Bind, name='getattr')
def getattr_(env, dtype, attr):
    return env.call('getattr', dtype.items.dtype, attr)


@ufunc.resolver(PgQueryBuilder, File, Bind, name='getattr')
def getattr_(env, dtype, attr):
    return ReservedProperty(dtype, attr.name)


@ufunc.resolver(PgQueryBuilder, Ref, Bind, name='getattr')
def getattr_(env, dtype, attr):
    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype.prop, prop)


@ufunc.resolver(PgQueryBuilder, BackRef, Bind, name='getattr')
def getattr_(env, dtype, attr):
    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype.prop, prop)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Bind, name='getattr')
def getattr_(env, fpr, attr):
    prop = fpr.right.dtype.model.properties[attr.name]
    return ForeignProperty(fpr, fpr.right, prop)


@ufunc.resolver(PgQueryBuilder, Inherit, Bind, name='getattr')
def getattr_(env, dtype, attr):
    return InheritForeignProperty(dtype.prop.model, attr.name, dtype.prop)


@ufunc.resolver(PgQueryBuilder, ExternalRef, Bind, name='getattr')
def getattr_(env, dtype, attr):
    return dtype


@ufunc.resolver(PgQueryBuilder, Text, Bind, name='getattr')
def getattr_(env, dtype, bind):
    if dtype.prop.name in env.model.properties:
        prop = env.model.properties[dtype.prop.name]
        if bind.name in prop.dtype.langs:
            return prop.dtype.langs[bind.name].dtype
        raise LangNotDeclared(dtype, lang=bind.name)


@ufunc.resolver(PgQueryBuilder, Expr)
def select(env, expr):
    keys = [str(k) for k in expr.args]
    args, kwargs = expr.resolve(env)
    args = list(zip(keys, args)) + list(kwargs.items())

    env.select = {}

    if args:
        for key, arg in args:
            selected = env.call('select', arg)
            if selected is not None:
                env.select[key] = selected
    else:
        env.call('select', Star())

    if not (len(args) == 1 and args[0][0] == '_page'):
        assert env.select, args


@ufunc.resolver(PgQueryBuilder, Star)
def select(env, arg: Star) -> None:
    for prop in take(env.model.properties).values():
        # if authorized(env.context, prop, (Action.GETALL, Action.SEARCH)):
        # TODO: This line above should come from a getall(request),
        #       because getall can be used internally for example for
        #       writes.
        env.select[prop.place] = env.call('select', prop.dtype)


@ufunc.resolver(PgQueryBuilder, Bind)
def select(env, arg):
    if arg.name == '_type':
        return Selected(None, env.model.properties['_type'])
    if arg.name == '_page':
        return None
    prop = _get_property_for_select(env, arg.name)
    return env.call('select', prop.dtype)


def _get_property_for_select(env: PgQueryBuilder, name: str):
    prop = env.model.properties.get(name)
    if prop and authorized(env.context, prop, Action.SEARCH):
        return prop
    else:
        raise FieldNotInResource(env.model, property=name)


@dataclasses.dataclass
class Selected:
    item: Any
    prop: Property = None

    def __eq__(self, other):
        if isinstance(other, Selected):
            return self.prop == other.prop
        return False


@ufunc.resolver(PgQueryBuilder, DataType)
def select(env, dtype):
    table = env.backend.get_table(env.model)

    if dtype.prop.list is None:
        column = env.backend.get_column(table, dtype.prop, select=True)
    else:
        # XXX: Probably if dtype.prop is in a nested list, we need to get the
        #      first list. Because if I remember correctly, only first list is
        #      stored on the main table.
        column = env.backend.get_column(table, dtype.prop.list, select=True)
    return Selected(column, dtype.prop)


@ufunc.resolver(PgQueryBuilder, String)
def select(env, dtype):
    env.call('validate_dtype_for_select', dtype, _gather_selected_properties(env))
    if dtype.prop.list is None:
        column = env.backend.get_column(env.table, dtype.prop)
    else:
        column = env.backend.get_column(env.table, dtype.prop.list)
    return Selected(column, dtype.prop)


@ufunc.resolver(PgQueryBuilder, Text)
def select(env, dtype):
    env.call('validate_dtype_for_select', dtype, _gather_selected_properties(env))
    if dtype.prop.list is None:
        column = _get_column_with_extra(env, dtype.prop)
    else:
        column = env.backend.get_column(env.table, dtype.prop.list)
    return Selected(column, dtype.prop)


@ufunc.resolver(PgQueryBuilder, Object)
def select(env, dtype):
    columns = []
    for prop in take(dtype.properties).values():
        sel = env.call('select', prop.dtype)
        if sel is not None:
            if isinstance(sel.item, list):
                columns += sel.item
            else:
                columns += [sel.item]
    return Selected(columns, dtype.prop)


@ufunc.resolver(PgQueryBuilder, File)
def select(env, dtype):
    table = env.backend.get_table(env.model)
    columns = [
        table.c[dtype.prop.place + '._id'],
        table.c[dtype.prop.place + '._content_type'],
        table.c[dtype.prop.place + '._size'],
    ]
    same_backend = dtype.backend.name == dtype.prop.model.backend.name
    if same_backend or BackendFeatures.FILE_BLOCKS in dtype.backend.features:
        columns += [
            table.c[dtype.prop.place + '._bsize'],
            table.c[dtype.prop.place + '._blocks'],
        ]
    return Selected(columns, dtype.prop)


@ufunc.resolver(PgQueryBuilder, ReservedProperty)
def select(env, prop):
    return env.call('select', prop.dtype, prop.param)


@ufunc.resolver(PgQueryBuilder, File, str)
def select(env, dtype, leaf):
    table = env.backend.get_table(env.model)
    if leaf == '_content':
        return env.call('select', dtype)
    else:
        column = table.c[dtype.prop.place + '.' + leaf]
        return Selected(column, dtype.prop)


@ufunc.resolver(PgQueryBuilder, Ref)
def select(env, dtype):
    uri = dtype.model.uri_prop
    if env.query_params.prioritize_uri and uri is not None:
        fpr = ForeignProperty(None, dtype.prop, dtype.model.properties['_id'])
        table = env.get_joined_table(fpr)
        column = table.c[uri.place]
        column = column.label(dtype.prop.place + '._uri')
    else:
        table = env.backend.get_table(env.model)
        column = table.c[dtype.prop.place + '._id']
    return Selected(column, dtype.prop)


@ufunc.resolver(PgQueryBuilder, ExternalRef)
def select(env, dtype):
    table = env.backend.get_table(env.model)
    columns = []
    if dtype.model.given.pkeys or dtype.explicit:
        props = dtype.refprops
    else:
        props = [dtype.model.properties['_id']]
    for prop in props:
        column = table.c[f"{dtype.prop.place}.{prop.place}"]
        columns.append(column)
    return Selected(columns, dtype.prop)


@ufunc.resolver(PgQueryBuilder, BackRef)
def select(env, dtype):
    fpr = ForeignProperty(
        None,
        left=dtype.prop,
        right=dtype.refprop,
    )
    refprop = dtype.refprop
    required_columns = []
    return_columns = []
    if refprop.level is None or refprop.level > Level.open:
        column_name = fpr.right.model.properties['_id'].name
        label = f'{dtype.prop.name}.{column_name}'
        required_columns.append((column_name, label))
        return_columns.append(label)
    else:
        required_columns = []
        for prop in dtype.refprop.dtype.refprops:
            column_name = prop.name
            label = f'{dtype.prop.name}.{column_name}'
            required_columns.append((column_name, label))
            return_columns.append(label)
    selector = env.generate_backref_select(dtype.prop, dtype.refprop, required_columns, False)
    table = env.get_backref_joined_table(fpr, selector)
    if len(return_columns) == 1:
        column = table.c[return_columns[0]]
        return Selected(column, fpr.left)
    else:
        collected_columns = []
        for col in return_columns:
            collected_columns.append(table.c[col])
        return Selected(collected_columns, fpr.left)


@ufunc.resolver(PgQueryBuilder, ArrayBackRef)
def select(env, dtype):
    fpr = ForeignProperty(
        None,
        left=dtype.prop,
        right=dtype.refprop,
    )
    refprop = dtype.refprop
    required_columns = []
    if refprop.level is None or refprop.level > Level.open:
        column_name = fpr.right.model.properties['_id'].name
        required_columns.append((column_name, column_name))
    else:
        required_columns = []
        for prop in dtype.refprop.dtype.refprops:
            column_name = prop.name
            required_columns.append((column_name, column_name))
    selector = env.generate_backref_select(dtype.prop, dtype.refprop, required_columns, True)
    table = env.get_backref_joined_table(fpr, selector)
    column = table.c[dtype.prop.name]
    return Selected(column, fpr.left)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def select(env: PgQueryBuilder, fpr: ForeignProperty):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    column = column.label(fpr.place)
    return Selected(column, fpr.right)


@ufunc.resolver(PgQueryBuilder, Inherit)
def select(env, dtype):
    table = env.get_joined_base_table(dtype.prop.model, dtype.prop.name)
    column = table.c[dtype.prop.name]
    column = column.label(dtype.prop.name)
    return Selected(column, dtype.prop)


@ufunc.resolver(PgQueryBuilder, InheritForeignProperty)
def select(env, dtype):
    table = env.get_joined_base_table(dtype.model, dtype.prop_name)
    column = table.c[dtype.prop_name]
    column = column.label(f"{dtype.base_prop.name}.{dtype.prop_name}")
    return Selected(column, dtype.base_prop)


@ufunc.resolver(PgQueryBuilder, Page)
def select(env, page):
    return_selected = []
    for item in page.by.values():
        selected = env.call('select', item.prop.dtype)
        return_selected.append(selected)
    return return_selected


@ufunc.resolver(PgQueryBuilder, int)
def limit(env, n):
    env.limit = n


@ufunc.resolver(PgQueryBuilder, int)
def offset(env, n):
    env.offset = n


@ufunc.resolver(PgQueryBuilder, Expr, name='and')
def and_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    if len(args) > 1:
        return sa.and_(*args)
    elif args:
        return args[0]


@ufunc.resolver(PgQueryBuilder, Expr, name='or')
def or_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    return env.call('or', args)


@ufunc.resolver(PgQueryBuilder, list, name='or')
def or_(env, args):
    if len(args) > 1:
        return sa.or_(*args)
    elif args:
        return args[0]


@ufunc.resolver(PgQueryBuilder)
def count(env):
    env.aggregate = True
    env.select = {
        'count()': Selected(sa.func.count().label('count()')),
    }


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]

COMPARE_EQUATIONS = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
]

COMPARE_STRING = [
    'eq',
    'startswith',
    'contains',
]


@ufunc.resolver(PgQueryBuilder, Bind, object, names=COMPARE)
def compare(env, op, field, value):
    prop = _get_from_flatprops(env.model, field.name)
    return env.call(op, prop.dtype, value)


def _get_from_flatprops(model: Model, prop: str):
    if prop in model.flatprops:
        return model.flatprops[prop]
    else:
        raise exceptions.FieldNotInResource(model, property=prop)


@ufunc.resolver(PgQueryBuilder, PrimaryKey, object, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, object, names=COMPARE)
def compare(env, op: str, fpr: ForeignProperty, value: Any):
    return env.call(op, fpr, fpr.right.dtype, value)


@ufunc.resolver(PgQueryBuilder, PrimaryKey, object, names=COMPARE)
def compare(env, op, dtype, value):
    value = str(value)
    try:
        uuid.UUID(value)
    except ValueError:
        raise exceptions.InvalidValue(dtype, op=op, arg=type(value).__name__)

    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, String, str, names=COMPARE)
def compare(env, op, dtype, value):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, (Integer, Number), (int, float), names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, DateTime, str, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.datetime.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, DateTime, datetime.datetime, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, Date, str, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.date.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, Date, datetime.date, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, Time, str, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.time.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, Time, datetime.time, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, DataType, object, names=COMPARE)
def compare(env, op, dtype, value):
    raise exceptions.InvalidValue(dtype, op=op, arg=type(value).__name__)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType, object, names=COMPARE)
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: DataType,
    value: Any,
):
    raise exceptions.InvalidValue(dtype, op=op, arg=type(value).__name__)


@ufunc.resolver(PgQueryBuilder, DataType, type(None))
def eq(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, Text, (str, Bind, type(None)))
def eq(env, dtype, value):
    column = _get_column_with_extra(env, dtype.prop)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType, type(None))
def eq(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: type(None),
):
    table = env.get_joined_table(fpr)
    column = env.backend.get_column(table, fpr.right)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, PrimaryKey, str)
def eq(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: type(None),
):
    table = env.backend.get_table(fpr.left.model)
    column = env.backend.get_column(table, fpr.left)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


def _ensure_non_empty(op, s):
    if s == '':
        raise EmptyStringSearch(op=op)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, String, str, names=COMPARE_STRING)
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: String,
    value: str,
):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, PrimaryKey, str, names=COMPARE_STRING)
def compare(env, op, dtype, value):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    column = env.backend.get_column(env.table, dtype.prop)
    return _sa_compare(op, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, (Integer, Number), (int, float), names=COMPARE_EQUATIONS)
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: Union[Integer, Number],
    value: Union[int, float],
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DateTime, str, names=COMPARE_EQUATIONS)
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: DateTime,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    value = datetime.datetime.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Date, str, names=COMPARE_EQUATIONS)
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: Date,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    value = datetime.date.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, String)
def lower(env, dtype):
    return Lower(dtype)


@ufunc.resolver(PgQueryBuilder, Recurse)
def lower(env, recurse):
    return Recurse([env.call('lower', arg) for arg in recurse.args])


@ufunc.resolver(PgQueryBuilder, Lower, str, names=COMPARE_STRING)
def compare(env, op, fn, value):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    column = _get_column_with_extra(env, fn.dtype.prop)
    column = sa.func.lower(column)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, fn.dtype.prop, cond)


def _sa_compare(op, column, value):
    # Convert JSONB value from -> to ->> with astext
    if isinstance(column.type, sa.JSON):
        if not isinstance(column, sa.Column):
            column = column.element
        column = column.astext

    if op == 'eq':
        return column == value

    if op == 'lt':
        return column < value

    if op == 'le':
        return column <= value

    if op == 'gt':
        return column > value

    if op == 'ge':
        return column >= value

    if op == 'contains':
        if isinstance(column.type, UUID):
            column = column.cast(sa.String)
        return column.contains(value)

    if op == 'startswith':
        if isinstance(column.type, UUID):
            column = column.cast(sa.String)
        return column.startswith(value)
    raise NotImplementedError


def _prepare_condition(env: PgQueryBuilder, prop: Property, cond):

    if prop.list is None:
        return cond

    main_table = env.table
    list_table = env.backend.get_table(prop.list, TableType.LIST)
    subqry = (
        sa.select(
            [list_table.c._rid],
            distinct=list_table.c._rid,
        ).
        where(cond).
        alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry,
        main_table.c._id == subqry.c._rid,
    )
    return subqry.c._rid.isnot(None)


@ufunc.resolver(PgQueryBuilder, DataType, type(None))
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType, type(None))
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: type(None),
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, String, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, Text, str)
def ne(env, dtype, value):
    column = _get_column_with_extra(env, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, String, str)
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, PrimaryKey, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, (Integer, Number), (int, float))
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, (Integer, Number), (int, float))
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: Union[Integer, Number],
    value: Union[int, float],
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, DateTime, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.datetime.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DateTime, str)
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DateTime,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    value = datetime.datetime.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, Date, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.date.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Date, str)
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: Date,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    value = datetime.date.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, Array, (object, type(None)), names=COMPARE)
def compare(env, op, dtype, value):
    return env.call(op, dtype.items.dtype, value)


@ufunc.resolver(PgQueryBuilder, Lower, str)
def ne(env, fn, value):
    column = _get_column_with_extra(env, fn.dtype.prop)
    column = sa.func.lower(column)
    return _ne_compare(env, fn.dtype.prop, column, value)


def _ne_compare(env: PgQueryBuilder, prop: Property, column, value):
    """Not equal operator is quite complicated thing and need explaining.

    If property is not defined within a list, just do `!=` comparison and be
    done with it.

    If property is in a list:

    - First check if there is at least one list item where field is not None
        (existance check).

    - Then check if there is no list items where field equals to given
        value.
    """

    if prop.list is None:
        return column != value

    main_table = env.backend.get_table(prop.model)
    list_table = env.backend.get_table(prop.list, TableType.LIST)

    # Check if at liest one value for field is defined
    subqry1 = (
        sa.select(
            [list_table.c._rid],
            distinct=list_table.c._rid,
        ).
        where(column != None).  # noqa
        alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry1,
        main_table.c._id == subqry1.c._rid,
    )

    # Check if given value exists
    subqry2 = (
        sa.select(
            [list_table.c._rid],
            distinct=list_table.c._rid,
        ).
        where(column == value).
        alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry2,
        main_table.c._id == subqry2.c._rid,
    )

    # If field exists and given value does not, then field is not equal to
    # value.
    return sa.and_(
        subqry1.c._rid != None,  # noqa
        subqry2.c._rid == None,
    )


FUNCS = [
    'lower',
    'upper',
]


@ufunc.resolver(PgQueryBuilder, Bind, names=FUNCS)
def func(env, name, field):
    prop = env.model.flatprops[field.name]
    return env.call(name, prop.dtype)


@ufunc.resolver(PgQueryBuilder, Bind)
def recurse(env, field):
    if field.name in env.model.leafprops:
        return Recurse([prop.dtype for prop in env.model.leafprops[field.name]])
    else:
        raise exceptions.FieldNotInResource(env.model, property=field.name)


@ufunc.resolver(PgQueryBuilder, Recurse, object, names=COMPARE)
def recurse(env, op, recurse, value):
    return env.call('or', [
        env.call(op, arg, value)
        for arg in recurse.args
    ])


@ufunc.resolver(PgQueryBuilder, Expr, name='any')
def any_(env, expr):
    args, kwargs = expr.resolve(env)
    op, field, *args = args
    if isinstance(op, Bind):
        op = op.name
    return env.call('or', [
        env.call(op, field, arg)
        for arg in args
    ])


@ufunc.resolver(PgQueryBuilder, Expr)
def sort(env, expr):
    args, kwargs = expr.resolve(env)
    env.sort = [
        env.call('sort', arg) for arg in args
    ]


@ufunc.resolver(PgQueryBuilder, Bind)
def sort(env, field):
    prop = _get_from_flatprops(env.model, field.name)
    return env.call('asc', prop.dtype)


@ufunc.resolver(PgQueryBuilder, DataType)
def sort(env, dtype):
    return env.call('asc', dtype)


@ufunc.resolver(PgQueryBuilder, Negative_)
def sort(env, field):
    prop = _get_from_flatprops(env.model, field.name)
    return env.call('desc', prop.dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def sort(env: PgQueryBuilder, fpr: ForeignProperty):
    return env.call('asc', fpr, fpr.right.dtype)


@ufunc.resolver(PgQueryBuilder, Positive)
def sort(env, sign):
    return env.call('asc', sign.arg)


@ufunc.resolver(PgQueryBuilder, Negative)
def sort(env, sign):
    return env.call('desc', sign.arg)


@ufunc.resolver(PgQueryBuilder, DataType)
def asc(env, dtype):
    column = _get_sort_column(env, dtype.prop)
    return column.asc()


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def asc(env: PgQueryBuilder, fpr: ForeignProperty):
    return env.call('asc', fpr, fpr.right.dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType)
def asc(env: PgQueryBuilder, fpr: ForeignProperty, dtype: DataType):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    return column.asc()


@ufunc.resolver(PgQueryBuilder, DataType)
def desc(env, dtype):
    column = _get_sort_column(env, dtype.prop)
    return column.desc()


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def desc(env: PgQueryBuilder, fpr: ForeignProperty):
    return env.call('desc', fpr, fpr.right.dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType)
def desc(env: PgQueryBuilder, fpr: ForeignProperty, dtype: DataType):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.place]
    return column.desc()


@ufunc.resolver(PgQueryBuilder, Array, names=['asc', 'desc'])
def sort(env, name, dtype: Array):
    return env.call(name, dtype.items.dtype)


def _get_column_with_extra(env: PgQueryBuilder, prop: Property):
    default_langs = env.context.get('config').languages
    column = env.backend.get_column(env.table, prop, langs=env.query_params.lang_priority, push=env.query_params.push, default_langs=default_langs)
    return column


def _get_sort_column(env: PgQueryBuilder, prop: Property):
    column = _get_column_with_extra(env, prop)

    if prop.list is None:
        return column

    main_table = env.table
    list_table = env.backend.get_table(prop.list, TableType.LIST)
    subqry = (
        sa.select(
            [list_table.c._rid, column.label('value')],
            distinct=list_table.c._rid,
        ).alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry,
        main_table.c._id == subqry.c._rid,
    )
    return subqry.c.value


@ufunc.resolver(PgQueryBuilder, Bind)
def negative(env, field) -> Negative:
    if field.name in env.model.properties:
        prop = env.model.properties[field.name]
    else:
        raise FieldNotInResource(env.model, property=field.name)
    return Negative(prop.dtype)


@ufunc.resolver(PgQueryBuilder, Bind)
def positive(env, field) -> Positive:
    if field.name in env.model.properties:
        prop = env.model.properties[field.name]
    else:
        raise FieldNotInResource(env.model, property=field.name)
    return Positive(prop.dtype)


@ufunc.resolver(PgQueryBuilder, DataType)
def negative(env, dtype) -> Negative:
    return Negative(dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def negative(env: PgQueryBuilder, fpr: ForeignProperty) -> Negative:
    return Negative(fpr)


@ufunc.resolver(PgQueryBuilder, DataType)
def positive(env, dtype) -> Positive:
    return Positive(dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def positive(env: PgQueryBuilder, fpr: ForeignProperty) -> Positive:
    return Positive(fpr)



from typing import AsyncIterator

from starlette.requests import Request

from spinta import commands
from spinta.types.datatype import File
from spinta.components import Context, Model, DataItem, Action, UrlParams
from spinta.backends.nobackend.components import NoBackend
from spinta.exceptions import BackendNotGiven


@commands.insert.register(Context, Model, NoBackend)
async def insert(
    context: Context,
    model: Model,
    backend: NoBackend,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    raise BackendNotGiven(model)


@commands.update.register(Context, Model, NoBackend)
async def update(
    context: Context,
    model: Model,
    backend: NoBackend,
    *,
    dstream: dict,
    stop_on_error: bool = True,
):
    raise BackendNotGiven(model)


@commands.delete.register(Context, Model, NoBackend)
async def delete(
    context: Context,
    model: Model,
    backend: NoBackend,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    raise BackendNotGiven(model)


@commands.push.register(Context, Request, File, NoBackend)
async def push(
    context: Context,
    request: Request,
    dtype: File,
    backend: NoBackend,
    *,
    action: Action,
    params: UrlParams,
):
    prop = dtype.prop
    raise BackendNotGiven(prop.model)



from __future__ import annotations

from typing import Iterable
from typing import TYPE_CHECKING
from typing import Union

from spinta.core.enums import Access
from spinta.utils.enums import enum_by_name, get_enum_by_name

if TYPE_CHECKING:
    from spinta.dimensions.enum.components import EnumItem
    from spinta.components import Model
    from spinta.components import Namespace
    from spinta.components import Property
    from spinta.datasets.components import Dataset
    from spinta.datasets.components import Resource
    from spinta.manifests.components import Manifest


def load_access_param(
    component: Union[
        Dataset,
        Resource,
        Namespace,
        Model,
        Property,
        EnumItem,
    ],
    given_access: str,
    parents: Iterable[Union[
        Manifest,
        Dataset,
        Namespace,
        Model,
        Property,
    ]] = (),
) -> None:
    access = enum_by_name(component, 'access', Access, given_access)

    # If child has higher access than parent, increase parent access.
    if access is not None:
        for parent in parents:
            if parent.access is None or access > parent.access:
                parent.access = access

    component.access = access
    component.given.access = given_access


def link_access_param(
    component: Union[
        Dataset,
        Resource,
        Namespace,
        Model,
        Property,
        EnumItem,
    ],
    parents: Iterable[Union[
        Manifest,
        Dataset,
        Namespace,
        Model,
        Property,
    ]] = (),
    *,
    use_given: bool = True,
) -> None:
    if component.access is None:
        for parent in parents:
            candidate = parent.given.access if use_given else parent.access
            if candidate:
                if isinstance(candidate, str):
                    candidate = get_enum_by_name(Access, candidate)
                component.access = candidate
                break
        else:
            component.access = Access.protected



import base64
import json

from spinta.utils.json import fix_data_for_json


def is_url_safe_base64(s):
    try:
        return base64.urlsafe_b64encode(base64.urlsafe_b64decode(s)) == s
    except Exception:
        return False


def encode_page_values(values: list):
    return base64.urlsafe_b64encode(json.dumps(fix_data_for_json(values)).encode('ascii'))



import asyncio
from typing import List
from typing import Optional

import click
from typer import Argument
from typer import Context as TyperContext

from spinta import commands
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.store import load_store
from spinta.cli.helpers.store import prepare_manifest
from spinta.core.context import configure_context


def bootstrap(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Manifest files to load"
    )),
):
    """Initialize backends

    This will create tables and sync manifest to backends.
    """
    context = configure_context(ctx.obj, manifests)
    store = prepare_manifest(context, ensure_config_dir=True)

    with context:
        require_auth(context)
        commands.bootstrap(context, store.manifest)


def sync(ctx: TyperContext):
    """Sync source manifests into main manifest

    Single main manifest can be populated from multiple different backends.
    """
    context = ctx.obj
    store = load_store(context)
    commands.load(context, store.internal, into=store.manifest)
    with context:
        require_auth(context)
        coro = commands.sync(context, store.manifest)
        asyncio.run(coro)
    click.echo("Done.")


def migrate(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Manifest files to load"
    )),
):
    """Migrate schema change to backends"""
    context = configure_context(ctx.obj, manifests)
    store = prepare_manifest(context, ensure_config_dir=True)
    with context:
        require_auth(context)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(commands.migrate(context, store.manifest))


def freeze(ctx: TyperContext):
    """Detect schema changes and create new version

    This will read current manifest structure, compare it with a previous
    freezed version and will generate new migration version if current and last
    versions do not match.
    """
    context = ctx.obj
    # Load just store, manifests will be loaded by freeze command.
    store = load_store(context)
    with context:
        require_auth(context)
        commands.freeze(context, store.manifest)
    click.echo("Done.")



from copy import copy

from spinta import commands
from spinta.components import Context
from spinta.exceptions import PartialTypeNotFound
from spinta.types.datatype import Partial, Ref


@commands.link.register(Context, Partial)
def link(context: Context, dtype: Partial):
    prop = dtype.prop
    if prop.parent:
        if isinstance(prop.parent.dtype, Ref):
            props = prop.dtype.properties
            prop.dtype = copy(prop.parent.dtype.model.properties[prop.name].dtype)
            prop.dtype.properties = props
            prop.given.name = ''

            for partial_prop in props.values():
                partial_prop.model = prop.parent.dtype.model
                commands.link(context, partial_prop)

        else:
            raise PartialTypeNotFound(dtype)
    else:
        raise PartialTypeNotFound(dtype)



from enum import Enum

from spinta.manifests.components import Manifest


class DictFormat(Enum):
    JSON = 'json'
    XML = 'xml'
    HTML = 'html'


class DictManifest(Manifest):
    format: DictFormat = None


class JsonManifest(DictManifest):
    type = 'json'
    format: DictFormat = DictFormat.JSON

    @staticmethod
    def detect_from_path(path: str) -> bool:
        return path.endswith('.json')


class XmlManifest(DictManifest):
    type = 'xml'
    format: DictFormat = DictFormat.XML

    @staticmethod
    def detect_from_path(path: str) -> bool:
        return path.endswith('.xml')



from typing import Optional, List

import io
import uuid

import sqlalchemy as sa


DEFAULT_BLOCK_SIZE = 8 << 20  # 8MB


class DatabaseFile(io.RawIOBase):

    def __init__(
        self,
        conn,
        table: sa.Table,
        size: int = 0,
        blocks: Optional[List[str]] = None,
        bsize: int = DEFAULT_BLOCK_SIZE,
        *,
        mode: str = 'r',
    ):
        self.size = size
        self.blocks = [] if blocks is None else blocks[:]
        self.bsize = bsize
        self.mode = mode
        self.dirty = False

        self._conn = conn
        self._table = table
        self._pos = 0
        self._block = None  # Currenly loaded block.
        self._block_id = None  # Currently loaded block id.

        if self.mode not in ('r', 'w', 'a'):
            raise Exception(f"Unsupported file mode {self.mode!r}.")

        if self.mode == 'a':
            self._pos = self.size

    def seekable(self):
        return True

    def readable(self):
        return self.mode == 'r'

    def writable(self):
        return self.mode in ('w', 'a')

    def tell(self):
        return self._pos

    def seek(self, pos, whence=0):
        offset = self._pos // self.bsize

        if whence == 0:
            self._pos = pos
        elif whence == 1:
            self._pos += pos
        elif whence == 2:
            self.size += pos
        else:
            raise Exception(f"Unsupported whence ({whence}).")

        self._check_position(self._pos)

        if offset != self._pos // self.bsize:
            self._block = None

        return self._pos

    def flush(self):
        self._checkClosed()

    def read(self, size=-1):

        if size is None or size < 0 or self._pos + size > self.size:
            size = self.size - self._pos

        # Read current block.
        #
        #   +--------|-------|-------+
        #   |012345670123456701234567|
        #   +--------|-------|-------+
        #      |   `--  size == 4
        #      `-- self._pos == 2
        #
        self._read_block()
        pos = self._pos % self.bsize
        reminder = self.bsize - pos
        if reminder >= size:
            self._pos += size
            return self._block[pos:pos + size]

        # Read full blocks.
        buffer = self._block[pos:]
        self._pos += reminder
        blocks, reminder = divmod(size - reminder, self.bsize)

        for i in range(blocks):
            self._read_block()
            buffer += self._block
            self._pos += self.bsize

        # Read reminder block.
        if reminder:
            self._read_block()
            buffer += self._block[:reminder]
            self._pos += reminder

        return buffer

    def write(self, b):
        self._read_block()

        size = len(b)
        pos = self._pos % self.bsize
        reminder = self.bsize - pos

        if reminder >= size:
            self._block = self._block[:pos] + b + self._block[pos + size:]
            self._write_block()
            self._pos += size
            if self._pos > self.size:
                self.size = self._pos
            return size

        self._block = self._block[:pos] + b[:reminder]
        self._write_block()
        self._pos += reminder
        pos = reminder
        blocks, reminder = divmod(size - reminder, self.bsize)

        for i in range(blocks):
            self._block = b[pos:pos + self.bsize]
            self._write_block()
            self._pos += self.bsize
            pos += self.bsize

        if reminder:
            self._read_block()
            self._block = b[pos:] + self._block[len(b[pos:]):]
            self._write_block()
            self._pos += reminder

        if self._pos > self.size:
            self.size = self._pos

        return size

    def _read_block(self):
        block_ix = self._pos // self.bsize
        block_count = len(self.blocks)

        if block_ix >= block_count:
            self._block = b''
            self._block_id = None
            return

        if self._block is not None and self.blocks[block_ix] == self._block_id:
            # This block is already loaded.
            return

        # Load block from database.
        query = (
            sa.select([self._table.c._block]).
            where(self._table.c._id == self.blocks[block_ix])
        )

        self._block = self._conn.execute(query).scalar()
        self._block_id = self.blocks[block_ix]

        if self._block is None:
            raise Exception(
                f"Tried to access block {self._block_id} on "
                f"{self._table.name!r}, but it does not exist."
            )

        block_size = len(self._block)
        last_block = block_ix == block_count - 1
        if not last_block and block_size != self.bsize:
            raise Exception(
                f"Expected block size {self.bsize}, received {block_size}. "
                f"Block id: {self._block_id} on {self._table.name!r}."
            )

    def _write_block(self):
        block_ix = self._pos // self.bsize
        block_count = len(self.blocks)

        if block_ix >= block_count + 1:
            raise Exception(
                f"Tried to write block {block_ix}, but number of available "
                f"blocks is {block_count}."
            )

        if len(self._block) > self.bsize:
            block_size = len(self._block)
            raise Exception(
                f"Expected block size {self.bsize}, received {block_size}. "
                f"Block id: {self._block_id} on {self._table.name!r}."
            )

        query = self._table.insert().values(
            _id=str(uuid.uuid4()),
            _block=self._block,
        )
        result = self._conn.execute(query)
        block_id = result.inserted_primary_key[0]

        if block_ix >= block_count:
            self.blocks.append(block_id)
        else:
            self.blocks[block_ix] = block_id

        self.dirty = True
        self._block_id = block_id

    def _check_position(self, pos):
        if pos < 0:
            raise Exception(
                f"Attempt to seek to a negative position ({self._pos})."
            )

        if pos > self.size:
            raise Exception(
                f"Attempt to seek outside of file boundaries, attempted "
                f"position ({self._pos}), file size ({self.size})."
            )

    def __enter__(self):
        self._checkClosed()
        return self

    def __exit__(self, *args):
        self.close()



import datetime
import hashlib
import textwrap
import uuid

import itertools
import json
import logging
import pathlib
from copy import deepcopy, copy

import pprintpp
import time
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypedDict
from typing import NamedTuple

import msgpack
import requests
import sqlalchemy as sa
import tqdm
from requests import HTTPError
from typer import Argument
from typer import Context as TyperContext
from typer import Option
from typer import Exit
from typer import echo

from spinta import exceptions
from spinta import spyna
from spinta.auth import authorized
from spinta.auth import get_client_id_from_name, get_clients_path
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.data import ModelRow, count_rows, read_model_data, filter_allowed_props_for_model, \
    filter_dict_by_keys
from spinta.cli.helpers.data import ensure_data_dir
from spinta.cli.helpers.errors import ErrorCounter
from spinta.cli.helpers.store import prepare_manifest
from spinta.client import get_access_token
from spinta.client import get_client_credentials
from spinta.commands.read import get_page
from spinta.components import Action, Page
from spinta.components import Config
from spinta.components import Context
from spinta.components import Mode
from spinta.components import Model
from spinta.components import Store
from spinta.core.context import configure_context
from spinta.core.ufuncs import Expr
from spinta.datasets.enums import Level
from spinta.datasets.keymaps.synchronize import sync_keymap
from spinta.exceptions import InfiniteLoopWithPagination, UnauthorizedPropertyPush
from spinta.manifests.components import Manifest
from spinta.types.datatype import Ref
from spinta.types.namespace import sort_models_by_ref_and_base
from spinta.ufuncs.basequerybuilder.components import QueryParams
from spinta.ufuncs.basequerybuilder.ufuncs import filter_page_values
from spinta.utils.data import take
from spinta.utils.itertools import peek
from spinta.utils.json import fix_data_for_json
from spinta.utils.nestedstruct import flatten
from spinta.utils.units import tobytes
from spinta.utils.units import toseconds
from spinta.utils.sqlite import migrate_table

log = logging.getLogger(__name__)


def push(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Source manifest files to copy from"
    )),
    output: Optional[str] = Option(None, '-o', '--output', help=(
        "Output data to a given location, by default outputs to stdout"
    )),
    credentials: str = Option(None, '--credentials', help=(
        "Credentials file, defaults to {config_path}/credentials.cfg"
    )),
    dataset: str = Option(None, '-d', '--dataset', help=(
        "Push only specified dataset"
    )),
    auth: str = Option(None, '-a', '--auth', help=(
        "Authorize as a client, defaults to {default_auth_client}"
    )),
    limit: int = Option(None, help=(
        "Limit number of rows read from each model"
    )),
    chunk_size: str = Option('1m', help=(
        "Push data in chunks (1b, 1k, 2m, ...), default: 1m"
    )),
    stop_time: str = Option(None, help=(
        "Stop pushing after given time (1s, 1m, 2h, ...), by default does not "
        "stops until all data is pushed"
    )),
    stop_row: int = Option(None, help=(
        "Stop after pushing n rows, by default does not stop until all data "
        "is pushed"
    )),
    state: pathlib.Path = Option(None, help=(
        "Save push state into a file, by default state is saved to "
        "{data_path}/push/{remote}.db SQLite database file"
    )),
    mode: Mode = Option('external', help=(
        "Mode of backend operation, default: external"
    )),
    dry_run: bool = Option(False, '--dry-run', help=(
        "Read data to be pushed, but do not push or write data to the "
        "destination."
    )),
    stop_on_error: bool = Option(False, '--stop-on-error', help=(
        "Exit immediately on first error."
    )),
    no_progress_bar: bool = Option(False, '--no-progress-bar', help=(
        "Skip counting total rows to improve performance."
    )),
    retry_count: int = Option(5, '--retries', help=(
        "Repeat push until this count if there are errors."
    )),
    max_error_count: int = Option(50, '--max-errors', help=(
        "If errors exceed given number, push command will be stopped."
    )),
    incremental: bool = Option(False, '-i', '--incremental', help=(
        "Do an incremental push, only pushing objects from last page."
    )),
    page: Optional[List[str]] = Option(None, '--page', help=(
        "Page value from which rows will be pushed."
    )),
    page_model: str = Option(None, '--model', help=(
        "Model of the page value."
    )),
    synchronize: bool = Option(False, '--sync', help=(
        "Update push sync state, in {data_path}/push/{remote}.db"
    )),
):
    """Push data to external data store"""
    if chunk_size:
        chunk_size = tobytes(chunk_size)

    if stop_time:
        stop_time = toseconds(stop_time)

    context = configure_context(ctx.obj, manifests, mode=mode)
    store = prepare_manifest(context)
    config: Config = context.get('config')

    if credentials:
        credsfile = pathlib.Path(credentials)
        if not credsfile.exists():
            echo(f"Credentials file {credsfile} does not exit.")
            raise Exit(code=1)
    else:
        credsfile = config.credentials_file
    # TODO: Read client credentials only if a Spinta URL is given.
    creds = get_client_credentials(credsfile, output)

    if not state:
        ensure_data_dir(config.data_path / 'push')
        state = config.data_path / 'push' / f'{creds.remote}.db'

    state = f'sqlite:///{state}'

    manifest = store.manifest
    if dataset and dataset not in manifest.datasets:
        echo(str(exceptions.NodeNotFound(manifest, type='dataset', name=dataset)))
        raise Exit(code=1)

    ns = manifest.namespaces['']

    echo(f"Get access token from {creds.server}")
    token = get_access_token(creds)

    client = requests.Session()
    client.headers['Content-Type'] = 'application/json'
    client.headers['Authorization'] = f'Bearer {token}'

    with context:
        auth_client = auth or config.default_auth_client
        auth_client = get_client_id_from_name(get_clients_path(config), auth_client)
        require_auth(context, auth_client)

        _attach_backends(context, store, manifest)
        _attach_keymaps(context, store)
        error_counter = ErrorCounter(max_count=max_error_count)

        from spinta.types.namespace import traverse_ns_models

        models = traverse_ns_models(context, ns, Action.SEARCH, dataset, source_check=True)
        models = sort_models_by_ref_and_base(list(models))

        if state:
            state = _State(*_init_push_state(state, models))
            context.attach('push.state.conn', state.engine.begin)

        with manifest.keymap as km:
            first_time = km.first_time_sync()
            if first_time:
                synchronize = True
            dependant_models = extract_dependant_nodes(context, models, not synchronize)
            sync_keymap(
                context=context,
                keymap=km,
                client=client,
                server=creds.server,
                models=dependant_models,
                error_counter=error_counter,
                no_progress_bar=no_progress_bar,
                reset_cid=synchronize
            )
        _update_page_values_for_models(context, state.metadata, models, incremental, page_model, page)

        rows = _read_rows(
            context,
            client,
            creds.server,
            models,
            state,
            limit,
            stop_on_error=stop_on_error,
            retry_count=retry_count,
            no_progress_bar=no_progress_bar,
            error_counter=error_counter,
        )

        _push(
            context,
            client,
            creds.server,
            models,
            rows,
            state=state,
            stop_time=stop_time,
            stop_row=stop_row,
            chunk_size=chunk_size,
            dry_run=dry_run,
            stop_on_error=stop_on_error,
            error_counter=error_counter,
        )

        if error_counter.has_errors():
            raise Exit(code=1)


def extract_dependant_nodes(context: Context, models: List[Model], filter_pushed: bool):
    extracted_models = [] if filter_pushed else models.copy()
    for model in models:
        if model.base:
            if not model.base.level or model.base.level > Level.open:
                if model.base.parent not in extracted_models:
                    if filter_pushed and not (model.base.parent in models):
                        extracted_models.append(model.base.parent)
                    elif not filter_pushed:
                        extracted_models.append(model.base.parent)

        for prop in model.properties.values():
            if isinstance(prop.dtype, Ref):
                if authorized(context, prop, action=Action.SEARCH):
                    if not prop.level or prop.level > Level.open:
                        if prop.dtype.model not in extracted_models:
                            if filter_pushed and not (prop.dtype.model in models):
                                extracted_models.append(prop.dtype.model)
                            elif not filter_pushed:
                                extracted_models.append(prop.dtype.model)

    return extracted_models


def _update_page_values_for_models(context: Context, metadata: sa.MetaData, models: List[Model], incremental: bool, page_model: str, page: list):
    conn = context.get('push.state.conn')
    table = metadata.tables['_page']
    for model in models:
        if model.page.is_enabled:
            if incremental:
                if (
                    page and
                    page_model and
                    page_model == model.name
                ):
                    model.page.update_values_from_list(page)
                else:
                    page = conn.execute(
                        sa.select([table.c.value]).
                        where(
                            table.c.model == model.name
                        )
                    ).scalar()
                    values = json.loads(page) if page else {}
                    _load_page_from_dict(model, values)


class _State(NamedTuple):
    engine: sa.engine.Engine
    metadata: sa.MetaData


class _PushRow:
    model: Optional[Model]
    data: Dict[str, Any]
    # SHA1 checksum of data generated with _get_data_rev.
    checksum: Optional[str]
    # True if data has already been sent.
    saved: bool = False
    # If push request received an error
    error: bool = False
    # Row operation
    op: Optional[str] = None
    # If True, need to push data immediately
    push: bool = False
    # If True, need to include in request
    send: bool = True

    def __init__(
        self,
        model: Optional[Model],
        data: Dict[str, Any],
        checksum: str = None,
        saved: bool = False,
        error: bool = False,
        op: str = None,
        push: bool = False,
        send: bool = True,
    ):
        self.model = model
        self.data = data
        self.checksum = checksum
        self.saved = saved
        self.error = error
        self.op = op
        self.push = push
        self.send = send


PUSH_NOW = _PushRow(
    model=None,
    data={
        '_type': None
    },
    push=True,
    send=False,
)


def _push(
    context: Context,
    client: requests.Session,
    server: str,  # https://example.com/
    models: List[Model],
    rows: Iterable[_PushRow],
    *,
    state: Optional[_State] = None,
    stop_time: Optional[int] = None,    # seconds
    stop_row: Optional[int] = None,     # stop aftern given number of rows
    chunk_size: Optional[int] = None,   # split into chunks of given size in bytes
    dry_run: bool = False,              # do not send or write anything
    stop_on_error: bool = False,        # raise error immediately
    error_counter: ErrorCounter = None,
) -> None:
    if stop_time:
        rows = _add_stop_time(rows, stop_time)

    if state:
        rows = _check_push_state(context, rows, state.metadata)

    rows = _prepare_rows_for_push(rows)

    if stop_row:
        rows = itertools.islice(rows, stop_row)

    rows = _push_to_remote_spinta(
        client,
        server,
        rows,
        chunk_size,
        dry_run=dry_run,
        error_counter=error_counter,
    )
    if state and not dry_run:
        rows = _save_push_state(context, rows, state.metadata)

    try:
        _push_rows(rows, stop_on_error, error_counter)
    except:
        raise
    finally:
        if state and not dry_run:
            _save_page_values(context, models, state.metadata)


def _read_rows(
    context: Context,
    client: requests.Session,
    server: str,
    models: List[Model],
    state: _State,
    limit: int = None,
    *,
    stop_on_error: bool = False,
    retry_count: int = 5,
    no_progress_bar: bool = False,
    error_counter: ErrorCounter = None,
) -> Iterator[_PushRow]:
    yield from _get_model_rows(
        context,
        models,
        state.metadata,
        limit,
        stop_on_error=stop_on_error,
        no_progress_bar=no_progress_bar,
        error_counter=error_counter,
    )

    yield PUSH_NOW
    yield from _get_deleted_rows(
        models,
        context,
        state.metadata,
        no_progress_bar=no_progress_bar,
    )

    for i in range(1, retry_count + 1):
        yield PUSH_NOW
        counts = (
            _get_rows_with_errors_counts(
                models,
                context,
                state.metadata
            )
        )
        total_count = sum(counts.values())

        if total_count > 0:
            yield from _get_rows_with_errors(
                client,
                server,
                models,
                context,
                state.metadata,
                counts,
                retry=i,
                no_progress_bar=no_progress_bar,
                error_counter=error_counter,
            )
        else:
            break


def _push_rows(
    rows: Iterator[_PushRow],
    stop_on_error: bool = False,
    error_counter: ErrorCounter = None,
) -> None:
    while True:
        try:
            next(rows)
        except StopIteration:
            break
        except:
            if stop_on_error:
                raise
            log.exception("Error while reading data.")
        if error_counter and error_counter.has_reached_max():
            break


def _attach_backends(context: Context, store: Store, manifest: Manifest) -> None:
    context.attach('transaction', manifest.backend.transaction)
    backends = set()
    for backend in store.backends.values():
        backends.add(backend.name)
        context.attach(f'transaction.{backend.name}', backend.begin)
    for backend in manifest.backends.values():
        backends.add(backend.name)
        context.attach(f'transaction.{backend.name}', backend.begin)
    for dataset_ in manifest.datasets.values():
        for resource in dataset_.resources.values():
            if resource.backend and resource.backend.name not in backends:
                backends.add(resource.backend.name)
                context.attach(f'transaction.{resource.backend.name}', resource.backend.begin)


def _attach_keymaps(context: Context, store: Store) -> None:
    for keymap in store.keymaps.values():
        context.attach(f'keymap.{keymap.name}', lambda: keymap)


def _get_model_rows(
    context: Context,
    models: List[Model],
    metadata: sa.MetaData,
    limit: int = None,
    *,
    stop_on_error: bool = False,
    no_progress_bar: bool = False,
    error_counter: ErrorCounter = None
) -> Iterator[_PushRow]:
    counts = (
        count_rows(
            context,
            models,
            limit,
            stop_on_error=stop_on_error,
            error_counter=error_counter,
        )
        if not no_progress_bar
        else {}
    )
    push_counter = None
    if not no_progress_bar:
        push_counter = tqdm.tqdm(desc='PUSH', ascii=True, total=sum(counts.values()))
    rows = _iter_model_rows(
        context,
        models,
        counts,
        metadata,
        limit,
        stop_on_error=stop_on_error,
        no_progress_bar=no_progress_bar,
        push_counter=push_counter,
    )
    for row in rows:
        yield row

    if push_counter is not None:
        push_counter.close()


def _iter_model_rows(
    context: Context,
    models: List[Model],
    counts: Dict[str, int],
    metadata: sa.MetaData,
    limit: int = None,
    *,
    stop_on_error: bool = False,
    no_progress_bar: bool = False,
    push_counter: tqdm.tqdm = None,
) -> Iterator[ModelRow]:
    params = QueryParams()
    params.push = True
    for model in models:

        model_push_counter = None
        if not no_progress_bar:
            count = counts.get(model.name)
            model_push_counter = tqdm.tqdm(desc=model.name, ascii=True, total=count, leave=False)

        if model.page and model.page.is_enabled and model.page.by:
            rows = _read_rows_by_pages(
                context,
                model,
                metadata,
                limit,
                stop_on_error,
                push_counter,
                model_push_counter,
                params=params
            )
            for row in rows:
                yield row
        else:
            stream = read_model_data(
                context,
                model,
                limit,
                stop_on_error,
                params=params
            )
            for item in stream:
                if push_counter:
                    push_counter.update(1)
                if model_push_counter:
                    model_push_counter.update(1)
                yield _PushRow(model, item)

        if model_push_counter is not None:
            model_push_counter.close()


def _authorize_model_properties(context: Context, model: Model):
    for prop in model.properties.values():
        if not authorized(context, prop, Action.UPSERT):
            raise UnauthorizedPropertyPush(prop)


def _read_model_data_by_page(
    context: Context,
    model: Model,
    model_page: Page,
    limit: int = None,
    stop_on_error: bool = False,
    params: QueryParams = None
) -> Iterable[Dict[str, Any]]:

    if limit is None:
        query = None
    else:
        query = Expr('limit', limit)

    stream = get_page(
        context,
        model,
        model.backend,
        model_page,
        query,
        limit,
        params=params
    )

    if stop_on_error:
        stream = peek(stream)
    else:
        try:
            stream = peek(stream)
        except Exception:
            log.exception(f"Error when reading data from model {model.name}")
            return

    prop_filer, needs_filtering = filter_allowed_props_for_model(model)
    for item in stream:
        if needs_filtering:
            item = filter_dict_by_keys(prop_filer, item)
        yield item


def _read_rows_by_pages(
    context: Context,
    model: Model,
    metadata: sa.MetaData,
    limit: int = None,
    stop_on_error: bool = False,
    push_counter: tqdm.tqdm = None,
    model_push_counter: tqdm.tqdm = None,
    params: QueryParams = None
) -> Iterator[_PushRow]:
    conn = context.get('push.state.conn')
    config = context.get('config')

    page_size = config.push_page_size
    size = model.page.size or page_size or 1000

    model_table = metadata.tables[model.name]
    state_rows = _get_state_rows(
        context,
        deepcopy(model.page),
        model_table,
        size
    )
    rows = _read_model_data_by_page(
        context,
        model,
        deepcopy(model.page),
        limit,
        stop_on_error,
        params
    )
    total_count = 0
    data_push_count = 0
    state_push_count = 0
    data_row = next(rows, None)
    state_row = next(state_rows, None)

    while True:
        if limit:
            if total_count >= limit:
                break
            total_count += 1

        update_counter = True

        if data_push_count >= size or state_push_count >= size:
            state_rows = _get_state_rows(
                context,
                deepcopy(model.page),
                model_table,
                size
            )
            rows = _read_model_data_by_page(
                context,
                model,
                deepcopy(model.page),
                limit,
                stop_on_error
            )

            data_push_count = 0
            state_push_count = 0
            data_row = next(rows, None)
            state_row = next(state_rows, None)

        if data_row is None and state_row is None:
            break

        if data_row and state_row:
            row = _PushRow(model, data_row)
            row.op = 'insert'
            row.checksum = _get_data_checksum(row.data)

            equals = _compare_data_with_state_rows(data_row, state_row, model_table)
            if equals:
                row.op = 'patch'
                row.saved = True
                row.data['_revision'] = state_row[model_table.c.revision]
                if state_row[model_table.c.checksum] != row.checksum or not _compare_data_with_state_row_keys(data_row, state_row, model_table, model.page):
                    yield row

                data_push_count += 1
                state_push_count += 1
                _update_model_page_with_new(model.page, model_table, data_row=data_row)
                data_row = next(rows, None)
                state_row = next(state_rows, None)

            else:
                delete_cond = _compare_for_delete_row(state_row, data_row, model_table, model.page)
                if delete_cond:
                    conn.execute(
                        sa.update(model_table).where(model_table.c.id == state_row["id"]).values(pushed=None)
                    )

                    state_push_count += 1
                    _update_model_page_with_new(model.page, model_table, state_row=state_row)
                    state_row = next(state_rows, None)
                    update_counter = False
                else:
                    yield row

                    data_push_count += 1
                    _update_model_page_with_new(model.page, model_table, data_row=data_row)
                    data_row = next(rows, None)
        else:
            if data_row:
                row = _PushRow(model, data_row)
                row.op = 'insert'
                row.checksum = _get_data_checksum(row.data)
                yield row

                data_push_count += 1
                _update_model_page_with_new(model.page, model_table, data_row=data_row)
                data_row = next(rows, None)
            else:
                conn.execute(
                    sa.update(model_table).where(model_table.c.id == state_row["id"]).values(pushed=None)
                )

                state_push_count += 1
                _update_model_page_with_new(model.page, model_table, state_row=state_row)
                state_row = next(state_rows, None)
                update_counter = False

        if update_counter:
            if push_counter:
                push_counter.update(1)
            if model_push_counter:
                model_push_counter.update(1)


def _update_model_page_with_new(
    model_page: Page,
    model_table: sa.Table,
    state_row: Any = None,
    data_row: Any = None,
):
    if state_row:
        for by, page_by in model_page.by.items():
            val = state_row[model_table.c[f"page.{page_by.prop.name}"]]
            model_page.update_value(by, page_by.prop, val)
    elif data_row:
        model_page.update_values_from_list(data_row['_page'])


def _compare_for_delete_row(
    state_row: Any,
    data_row: Any,
    model_table: sa.Table,
    page: Page,
):
    delete_cond = False

    if '_page' in data_row:
        decoded = data_row['_page']
        for i, (by, page_by) in enumerate(page.by.items()):
            state_val = state_row[model_table.c[f"page.{page_by.prop.name}"]]
            data_val = decoded[i]
            compare = True
            if state_val is not None and data_val is not None:
                if by.startswith('-'):
                    compare = state_val > data_val
                else:
                    compare = state_val < data_val
            if state_val != data_val:
                delete_cond = compare
                break
    else:
        delete_cond = True
    return delete_cond


def _compare_data_with_state_rows(
    data_row: Any,
    state_row: Any,
    model_table: sa.Table,
):
    return data_row["_id"] == state_row[model_table.c["id"]]


def _compare_data_with_state_row_keys(
    data_row: Any,
    state_row: Any,
    model_table: sa.Table,
    page: Page
):
    equals = True
    decoded = data_row.get("_page")
    for i, (by, page_by) in enumerate(page.by.items()):
        state_val = state_row[model_table.c[f"page.{page_by.prop.name}"]]
        data_val = decoded[i]
        if state_val != data_val:
            equals = False
            break
    return equals


def _get_state_rows(
    context: Context,
    model_page: Page,
    table: sa.Table,
    size: int
) -> sa.engine.LegacyCursorResult:
    conn = context.get('push.state.conn')

    order_by = []

    for by, page_by in model_page.by.items():
        if by.startswith('-'):
            order_by.append(sa.desc(table.c[f"page.{page_by.prop.name}"]))
        else:
            order_by.append(sa.asc(table.c[f"page.{page_by.prop.name}"]))
    last_value = None
    while True:
        finished = True
        from_cond = _construct_where_condition_from_page(model_page, table)

        if from_cond is not None:
            rows = conn.execute(
                sa.select([table]).
                where(
                    from_cond
                ).order_by(*order_by).limit(size)
            )
        else:
            rows = conn.execute(
                sa.select([table])
                .order_by(*order_by).limit(size)
            )
        first_value = None
        for row in rows:
            if finished:
                finished = False

            if first_value is None:
                first_value = row
                if first_value == last_value:
                    raise InfiniteLoopWithPagination()
                else:
                    last_value = first_value

            for by, page_by in model_page.by.items():
                model_page.update_value(by, page_by.prop, row[f"page.{page_by.prop.name}"])
            yield row

        if finished:
            break


def _prepare_rows_for_push(rows: Iterable[_PushRow]) -> Iterator[_PushRow]:
    for row in rows:
        row.data['_op'] = row.op
        if row.op == 'patch':
            where = {
                'name': 'eq',
                'args': [
                    {'name': 'bind', 'args': ['_id']},
                    row.data['_id'],
                ]
            }
            row.data['_where'] = spyna.unparse(where)
        elif row.op == 'insert':
            # if _revision is passed insert action gives ManagedProperty error
            if '_revision' in row.data:
                row.data.pop('_revision')
        elif row.op == 'delete' or not row.send:
            pass
        else:
            raise NotImplementedError(row.op)
        yield row


def _push_to_remote_spinta(
    client: requests.Session,
    server: str,
    rows: Iterable[_PushRow],
    chunk_size: int,
    *,
    dry_run: bool = False,
    stop_on_error: bool = False,
    error_counter: ErrorCounter = None,
) -> Iterator[_PushRow]:
    prefix = '{"_data":['
    suffix = ']}'
    slen = len(suffix)
    chunk = prefix
    ready: List[_PushRow] = []

    for row in rows:
        if error_counter:
            if error_counter.has_reached_max():
                break

        data = fix_data_for_json(row.data)
        tmp = data
        if '_page' in data:
            tmp = copy(data)
            tmp.pop('_page')
        data = json.dumps(tmp, ensure_ascii=False)

        if ready and (
            len(chunk) + len(data) + slen > chunk_size or
            row.push
        ):
            yield from _send_and_receive(
                client,
                server,
                ready,
                chunk + suffix,
                dry_run=dry_run,
                stop_on_error=stop_on_error,
                error_counter=error_counter,
            )
            chunk = prefix
            ready = []
        if not row.push and row.send:
            chunk += (',' if ready else '') + data
            ready.append(row)

    if ready:
        if error_counter:
            if not error_counter.has_reached_max():
                yield from _send_and_receive(
                    client,
                    server,
                    ready,
                    chunk + suffix,
                    dry_run=dry_run,
                    stop_on_error=stop_on_error,
                    error_counter=error_counter
                )
        else:
            yield from _send_and_receive(
                client,
                server,
                ready,
                chunk + suffix,
                dry_run=dry_run,
                stop_on_error=stop_on_error,
                error_counter=error_counter
            )


def _send_and_receive(
    client: requests.Session,
    server: str,
    rows: List[_PushRow],
    data: str,
    *,
    dry_run: bool = False,
    stop_on_error: bool = False,
    error_counter: ErrorCounter = None,
) -> Iterator[_PushRow]:
    if dry_run:
        recv = _send_data_dry_run(data)
    else:
        _, recv = _send_request(
            client,
            server,
            "POST",
            rows,
            data,
            stop_on_error=stop_on_error,
            error_counter=error_counter,
        )
        if recv:
            recv = recv['_data']
    yield from _map_sent_and_recv(rows, recv)


def _send_data_dry_run(
    data: str,
) -> Optional[List[Dict[str, Any]]]:
    """Pretend data has been sent to a target location."""
    recv = json.loads(data)['_data']
    for row in recv:
        if '_id' not in row:
            row['_id'] = str(uuid.uuid4())
        row['_rev'] = str(uuid.uuid4())
    return recv


def _send_request(
    client: requests.Session,
    server: str,
    method: str,
    rows: List[_PushRow],
    data: str,
    *,
    stop_on_error: bool = False,
    ignore_errors: Optional[List[int]] = None,
    error_counter: ErrorCounter = None,
) -> Tuple[Optional[int], Optional[Dict[str, Any]]]:
    data = data.encode('utf-8')
    if not ignore_errors:
        ignore_errors = []

    try:
        resp = client.request(method, server, data=data)
    except IOError as e:
        if error_counter:
            error_counter.increase()
        log.error(
            (
                "Error when sending and receiving data.%s\n"
                "Error: %s"
            ),
            _get_row_for_error(rows),
            e,
        )
        if stop_on_error:
            raise
        return None, None

    try:
        resp.raise_for_status()
    except HTTPError:
        if resp.status_code not in ignore_errors:
            if error_counter:
                error_counter.increase()
            try:
                recv = resp.json()
            except requests.JSONDecodeError:
                log.error(
                    (
                        "Error when sending and receiving data.\n"
                        "Server response (status=%s):\n%s"
                    ),
                    resp.status_code,
                    textwrap.indent(resp.text, '    '),
                )
                if stop_on_error:
                    raise
            else:
                errors = recv.get('errors')
                log.error(
                    (
                        "Error when sending and receiving data.%s\n"
                        "Server response (status=%s):\n%s"
                    ),
                    _get_row_for_error(rows, errors),
                    resp.status_code,
                    textwrap.indent(pprintpp.pformat(recv), '    '),
                )
            if stop_on_error:
                raise
        return resp.status_code, None

    return resp.status_code, resp.json()


class _ErrorContext(TypedDict):
    # Manifest name
    manifest: str

    # Dataset name
    dataset: str

    # Absolute model name (but does not start with /)
    model: str

    # Property name
    property: str

    # Property data type
    type: str

    # Model name in data source
    entity: str

    # Property name in data source
    attribute: str

    # Full class path of component involved in the error.
    component: str

    # Line number or other location (depends on manifest) node in manifest
    # involved in the error.
    schema: str

    # External object id
    id: str


class _Error(TypedDict):
    # Error type
    type: str

    # Error code (exception class name)
    code: str

    # Error context.
    context: _ErrorContext

    # Message template.
    tempalte: str

    # Message compiled from templates and context.
    message: str


def _get_row_for_error(
    rows: List[_PushRow],
    errors: Optional[List[_Error]] = None,
) -> str:
    message = []
    size = len(rows)

    if errors:
        rows_by_id = {
            row.data['_id']: row
            for row in rows
            if row.data and '_id' in row.data
        }
        for error in errors:
            ctx = error.get('context') or {}
            if 'id' in ctx and ctx['id'] in rows_by_id:
                row = rows_by_id[ctx['id']]
                message += [
                    f" Model {row.model.name}, data:",
                    ' ' + pprintpp.pformat(row.data),
                ]

    if len(message) == 0 and size > 0:
        row = rows[0]
        data = pprintpp.pformat(row.data)
        message += [
            (
                f" Model {row.model.name},"
                f" items in chunk: {size},"
                f" first item in chunk:"
            ),
            data
        ]

    return '\n'.join(message)


def _map_sent_and_recv(
    sent: List[_PushRow],
    recv: Optional[List[Dict[str, Any]]],
) -> Iterator[_PushRow]:
    if recv is None:
        # We don't have a response, because there was an error while
        # communicating with the target server.
        for row in sent:
            row.error = True
            yield row
    else:
        assert len(sent) == len(recv), (
            f"len(sent) = {len(sent)}, len(received) = {len(recv)}"
        )
        for sent_row, recv_row in zip(sent, recv):
            if '_id' in sent_row.data:
                _id = sent_row.data['_id']
            else:
                # after delete
                _id = spyna.parse(sent_row.data['_where'])['args'][1]

            assert _id == recv_row['_id'], (
                f"sent._id = {_id}, "
                f"received._id = {recv_row['_id']}"
            )
            sent_row.data['_revision'] = recv_row['_revision']
            sent_row.error = False
            yield sent_row


def _add_stop_time(
    rows: Iterable[_PushRow],
    stop: int,  # seconds
) -> Iterator[_PushRow]:
    start = time.time()
    for row in rows:
        yield row
        if time.time() - start > stop:
            break


def _init_push_state(
    dburi: str,
    models: List[Model],
) -> Tuple[sa.engine.Engine, sa.MetaData]:
    engine = sa.create_engine(dburi)
    metadata = sa.MetaData(engine)
    inspector = sa.inspect(engine)

    page_table = sa.Table(
        '_page', metadata,
        sa.Column('model', sa.Text, primary_key=True),
        sa.Column('property', sa.Text),
        sa.Column('value', sa.Text),
    )
    page_table.create(checkfirst=True)

    types = {
        'string': sa.Text,
        'date': sa.Date,
        'datetime': sa.DateTime,
        'time': sa.Time,
        'integer': sa.Integer,
        'number': sa.Numeric
    }

    for model in models:
        pagination_cols = []
        if model.page and model.page.by and model.backend.paginated:
            for page_by in model.page.by.values():
                _type = types.get(page_by.prop.dtype.name, sa.Text)
                pagination_cols.append(
                    sa.Column(f"page.{page_by.prop.name}", _type, index=True)
                )

        table = sa.Table(
            model.name, metadata,
            sa.Column('id', sa.Unicode, primary_key=True),
            sa.Column('checksum', sa.Unicode),
            sa.Column('revision', sa.Unicode),
            sa.Column('pushed', sa.DateTime),
            sa.Column('error', sa.Boolean),
            sa.Column('data', sa.Text),
            *pagination_cols
        )
        migrate_table(engine, metadata, inspector, table, renames={
            'rev': 'checksum',
        })

    return engine, metadata


def _get_model_type(row: _PushRow) -> str:
    return row.data['_type']


def _get_model(row: _PushRow) -> Model:
    return row.model


def _get_data_checksum(data: dict):
    data = fix_data_for_json(take(data))
    data = flatten([data])
    data = [[k, v] for x in data for k, v in sorted(x.items())]
    data = msgpack.dumps(data, strict_types=True)
    checksum = hashlib.sha1(data).hexdigest()
    return checksum


class _Saved(NamedTuple):
    revision: str
    checksum: str


def _reset_pushed(
    context: Context,
    models: List[Model],
    metadata: sa.MetaData,
):
    conn = context.get('push.state.conn')
    for model in models:
        table = metadata.tables[model.name]

        # reset pushed so we could see which objects were deleted
        conn.execute(
            table.update().
            values(pushed=None)
        )


def _check_push_state(
    context: Context,
    rows: Iterable[_PushRow],
    metadata: sa.MetaData
):
    conn = context.get('push.state.conn')

    for model_type, group in itertools.groupby(rows, key=_get_model_type):
        saved_rows = {}
        if model_type:
            table = metadata.tables[model_type]

            query = sa.select([table.c.id, table.c.revision, table.c.checksum])
            saved_rows = {
                state[table.c.id]: _Saved(
                    state[table.c.revision],
                    state[table.c.checksum],
                )
                for state in conn.execute(query)
            }

        for row in group:
            if row.send and not row.error and row.op != "delete":
                _id = row.data['_id']
                row.checksum = _get_data_checksum(row.data)
                saved = saved_rows.get(_id)
                if saved is None:
                    row.op = "insert"
                    row.saved = False
                else:
                    row.op = "patch"
                    row.saved = True
                    row.data['_revision'] = saved.revision
                    if saved.checksum == row.checksum:
                        conn.execute(
                            table.update().
                            where(table.c.id == _id).
                            values(
                                pushed=datetime.datetime.now()
                            )
                        )
                        # Skip if no changes, but only if not paginated, since pagination already skips those
                        if '_page' not in row.data:
                            continue

            yield row


def _save_push_state(
    context: Context,
    rows: Iterable[_PushRow],
    metadata: sa.MetaData,
) -> Iterator[_PushRow]:
    conn = context.get('push.state.conn')
    for row in rows:
        table = metadata.tables[row.data['_type']]

        if row.model.page and row.model.page.by and '_page' in row.data:
            loaded = row.data['_page']
            page = {
                f"page.{page_by.prop.name}": loaded[i]
                for i, page_by in enumerate(row.model.page.by.values())
            }
            row.data.pop('_page')
        else:
            page = {}

        if row.error and row.op != "delete":
            data = fix_data_for_json(row.data)
            data = json.dumps(data)
        else:
            data = None

        if '_id' in row.data:
            _id = row.data['_id']
        else:
            _id = spyna.parse(row.data['_where'])['args'][1]

        if row.op == "delete" and not row.error:
            conn.execute(
                table.delete().
                where(table.c.id == _id)
            )
        elif row.saved:
            conn.execute(
                table.update().
                where(table.c.id == _id).
                values(
                    id=_id,
                    revision=row.data.get('_revision'),
                    checksum=row.checksum,
                    pushed=datetime.datetime.now(),
                    error=row.error,
                    data=data,
                    **page
                )
            )
        else:
            conn.execute(
                table.insert().
                values(
                    id=_id,
                    revision=row.data.get('_revision'),
                    checksum=row.checksum,
                    pushed=datetime.datetime.now(),
                    error=row.error,
                    data=data,
                    **page
                )
            )
        yield row


def _save_page_values(
    context: Context,
    models: List[Model],
    metadata: sa.MetaData,
):
    conn = context.get('push.state.conn')
    page_table = metadata.tables['_page']

    for model in models:
        pagination_props = []
        saved = False
        if model.page and model.page.is_enabled and model.page.by:
            pagination_props = model.page.by.values()
            page_row = conn.execute(
                sa.select(page_table.c.model)
                .where(page_table.c.model == model.name)
            ).scalar()

            if page_row is not None:
                saved = True

        if pagination_props:
            value = {}
            for page_by in pagination_props:
                if page_by.value is not None:
                    value.update({
                        page_by.prop.name: page_by.value
                    })
            if value:
                value = json.dumps(value)
                if saved:
                    conn.execute(
                        page_table.update().
                        where(
                            (page_table.c.model == model.name)
                        ).
                        values(
                            value=value
                        )
                    )
                else:
                    conn.execute(
                        page_table.insert().
                        values(
                            model=model.name,
                            property=','.join([page_by.prop.name for page_by in pagination_props]),
                            value=value
                        )
                    )


def _get_deleted_rows(
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
    no_progress_bar: bool = False,
):
    counts = (
        _get_deleted_row_counts(
            models,
            context,
            metadata
        )
    )
    total_count = sum(counts.values())
    if total_count > 0:
        rows = _iter_deleted_rows(
            models,
            context,
            metadata,
            counts,
            no_progress_bar
        )
        if not no_progress_bar:
            rows = tqdm.tqdm(rows, 'PUSH DELETED', ascii=True, total=total_count)
        for row in rows:
            yield row


def _get_deleted_row_counts(
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
) -> dict:
    counts = {}
    conn = context.get('push.state.conn')
    for model in models:
        table = metadata.tables[model.name]
        required_condition = sa.and_(
            table.c.pushed.is_(None),
            table.c.error.is_(False)
        )

        if model.page and model.page.is_enabled and model.page.by:
            where_cond = _construct_where_condition_to_page(model.page, table)
        else:
            where_cond = None
        if where_cond is not None:
            where_cond = sa.and_(
                where_cond,
                required_condition
            )
        else:
            where_cond = required_condition

        row_count = conn.execute(
            sa.select(sa.func.count(table.c.id)).
            where(
                where_cond
            )
        )
        counts[model.name] = row_count.scalar()
    return counts


def _iter_deleted_rows(
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
    counts: Dict[str, int],
    no_progress_bar: bool = False,
) -> Iterable[_PushRow]:
    models = reversed(models)
    config = context.get('config')
    conn = context.get('push.state.conn')
    page_size = config.push_page_size
    for model in models:
        size = model.page.size or page_size or 1000
        table = metadata.tables[model.name]
        total = counts.get(model.name)

        if model.page and model.page.is_enabled and model.page.by:
            rows = _get_deleted_rows_with_page(context, model.page, table, size)
        else:
            rows = conn.execute(
                sa.select([table.c.id]).
                where(
                    table.c.pushed.is_(None) &
                    table.c.error.is_(False)
                )
            )
        if not no_progress_bar:
            rows = tqdm.tqdm(
                rows,
                model.name,
                ascii=True,
                total=total,
                leave=False
            )

        for row in rows:
            yield _prepare_deleted_row(model, row[table.c.id])


def _construct_where_condition_to_page(
    page: Page,
    table: sa.Table
):
    item_count = len(page.by.keys())
    where_list = []
    for i in range(item_count):
        where_list.append([])
    for i, (by, page_by) in enumerate(page.by.items()):
        if page_by.value:
            for n in range(item_count):
                if n >= i:
                    if n == i:
                        if n == item_count - 1:
                            if by.startswith('-'):
                                where_list[n].append(table.c[f"page.{page_by.prop.name}"] >= page_by.value)
                            else:
                                where_list[n].append(table.c[f"page.{page_by.prop.name}"] <= page_by.value)
                        else:
                            if by.startswith('-'):
                                where_list[n].append(table.c[f"page.{page_by.prop.name}"] > page_by.value)
                            else:
                                where_list[n].append(table.c[f"page.{page_by.prop.name}"] < page_by.value)
                    else:
                        where_list[n].append(table.c[f"page.{page_by.prop.name}"] == page_by.value)

    where_condition = None
    for where in where_list:
        condition = None
        for item in where:
            if condition is not None:
                condition = sa.and_(
                    condition,
                    item
                )
            else:
                condition = item
        if where_condition is not None:
            where_condition = sa.or_(
                where_condition,
                condition
            )
        else:
            where_condition = condition

    return where_condition


def _construct_where_condition_from_page(
    page: Page,
    table: sa.Table
):
    filtered = filter_page_values(page)
    item_count = len(filtered.by.keys())
    where_list = []
    for i in range(item_count):
        where_list.append([])
    for i, (by, page_by) in enumerate(filtered.by.items()):
        if page_by.value:
            for n in range(item_count):
                if n >= i:
                    if n == i:
                        if by.startswith('-'):
                            where_list[n].append(table.c[f"page.{page_by.prop.name}"] < page_by.value)
                        else:
                            where_list[n].append(table.c[f"page.{page_by.prop.name}"] > page_by.value)
                    else:
                        where_list[n].append(table.c[f"page.{page_by.prop.name}"] == page_by.value)

    where_condition = None
    for where in where_list:
        condition = None
        for item in where:
            if condition is not None:
                condition = sa.and_(
                    condition,
                    item
                )
            else:
                condition = item
        if where_condition is not None:
            where_condition = sa.or_(
                where_condition,
                condition
            )
        else:
            where_condition = condition

    return where_condition


def _get_deleted_rows_with_page(
    context: Context,
    model_page: Page,
    table: sa.Table,
    size: int,
) -> sa.engine.LegacyCursorResult:
    conn = context.get('push.state.conn')

    order_by = []

    for page_by in model_page.by.values():
        order_by.append(sa.asc(table.c[f"page.{page_by.prop.name}"]))

    from_page = Page()
    from_page.update_values_from_page(model_page)
    from_page.clear()

    required_where_cond = sa.and_(
        table.c.pushed.is_(None),
        table.c.error.is_(False)
    )
    last_value = None
    while True:
        finished = True
        from_cond = _construct_where_condition_from_page(from_page, table)
        to_cond = _construct_where_condition_to_page(model_page, table)

        where_cond = None
        if from_cond is not None:
            where_cond = from_cond
        if to_cond is not None:
            if where_cond is not None:
                sa.and_(
                    where_cond,
                    to_cond
                )
            else:
                where_cond = to_cond

        if where_cond is not None:
            where_cond = sa.and_(
                required_where_cond,
                where_cond
            )
        else:
            where_cond = required_where_cond

        deleted_rows = conn.execute(
            sa.select([table]).
            where(
                where_cond
            ).order_by(*order_by).limit(size)
        )
        first_value = None
        for row in deleted_rows:
            if finished:
                finished = False

            if first_value is None:
                first_value = row
                if first_value == last_value:
                    raise InfiniteLoopWithPagination()
                else:
                    last_value = first_value

            for by, page_by in from_page.by.items():
                from_page.update_value(by, page_by.prop, row[f"page.{page_by.prop.name}"])
            yield row

        if finished:
            break


def _prepare_deleted_row(
    model: Model,
    _id: str,
    error: bool = False
):
    where = {
        'name': 'eq',
        'args': [
            {'name': 'bind', 'args': ['_id']},
            _id,
        ]
    }
    return _PushRow(
        model,
        {
            '_type': model.name,
            '_where': spyna.unparse(where)
        },
        saved=True,
        op="delete",
        error=error
    )


def _get_rows_with_errors(
    client: requests.Session,
    server: str,
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
    counts: Dict[str, int],
    retry: int,
    no_progress_bar: bool = False,
    error_counter: ErrorCounter = None
):
    rows = _iter_rows_with_errors(
        client,
        server,
        models,
        context,
        metadata,
        counts,
        no_progress_bar,
        error_counter,
    )
    if not no_progress_bar:
        rows = tqdm.tqdm(rows, f'RETRY #{retry}', ascii=True, total=sum(counts.values()))
    for row in rows:
        yield row


def _get_rows_with_errors_counts(
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
) -> dict:
    counts = {}
    conn = context.get('push.state.conn')
    for model in models:
        table = metadata.tables[model.name]
        required_condition = sa.and_(
            table.c.error.is_(True)
        )
        if model.page and model.page.is_enabled and model.page.by:
            where_cond = _construct_where_condition_to_page(model.page, table)
        else:
            where_cond = None
        if where_cond is not None:
            where_cond = sa.and_(
                where_cond,
                required_condition
            )
        else:
            where_cond = required_condition

        row_count = conn.execute(
            sa.select(sa.func.count(table.c.id)).
            where(
                where_cond
            )
        )
        counts[model.name] = row_count.scalar()
    return counts


def _iter_rows_with_errors(
    client: requests.Session,
    server: str,
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
    counts: Dict[str, int],
    no_progress_bar: bool = False,
    error_counter: ErrorCounter = None,
) -> Iterable[ModelRow]:
    conn = context.get('push.state.conn')
    config = context.get('config')
    page_size = config.push_page_size

    for model in models:
        size = model.page.size or page_size or 1000
        table = metadata.tables[model.name]

        if model.page and model.page.is_enabled and model.page.by:
            rows = _get_deleted_rows_with_page(context, model.page, table, size)
        else:
            rows = conn.execute(
                sa.select([table.c.id, table.c.checksum, table.c.data]).
                where(
                    table.c.error.is_(True)
                )
            )
        if not no_progress_bar:
            rows = tqdm.tqdm(
                rows,
                model.name,
                ascii=True,
                total=counts.get(model.name),
                leave=False
            )

        yield from _prepare_rows_with_errors(
            client,
            server,
            context,
            rows,
            model,
            table,
            error_counter
        )


def _get_error_rows_with_page(
    context: Context,
    model_page: Page,
    table: sa.Table,
    size: int
):
    conn = context.get('push.state.conn')
    order_by = []

    for page_by in model_page.by.values():
        order_by.append(sa.asc(table.c[f"page.{page_by.prop.name}"]))

    from_page = Page()
    from_page.update_values_from_page(model_page)
    from_page.clear()

    required_where_cond = sa.and_(
        table.c.error.is_(True)
    )
    last_value = None
    while True:
        finished = True
        from_cond = _construct_where_condition_from_page(from_page, table)
        to_cond = _construct_where_condition_to_page(model_page, table)

        where_cond = None
        if from_cond is not None:
            where_cond = from_cond
        if to_cond is not None:
            if where_cond is not None:
                sa.and_(
                    where_cond,
                    to_cond
                )
            else:
                where_cond = to_cond

        if where_cond is not None:
            where_cond = sa.and_(
                required_where_cond,
                where_cond
            )
        else:
            where_cond = required_where_cond

        rows = conn.execute(
            sa.select([table]).
            where(
                where_cond
            ).order_by(*order_by).limit(size)
        )
        first_value = None
        for row in rows:
            if finished:
                finished = False

            if first_value is None:
                first_value = row
                if first_value == last_value:
                    raise InfiniteLoopWithPagination()
                else:
                    last_value = first_value

            for by, page_by in from_page.by.items():
                from_page.update_value(by, page_by.prop, row[f"page.{page_by.prop.name}"])
            yield row

        if finished:
            break


def _prepare_rows_with_errors(
    client: requests.Session,
    server: str,
    context: Context,
    rows,
    model: Model,
    table: sa.Table,
    error_counter: ErrorCounter = None
) -> Iterable[ModelRow]:
    conn = context.get('push.state.conn')
    for row in rows:
        type = model.model_type()
        _id = row[table.c.id]
        checksum = row[table.c.checksum]
        data = json.loads(row[table.c.data]) if row[table.c.data] else {}

        status_code, resp = _send_request(
            client,
            f'{server}/{type}/{_id}',
            "GET",
            rows=[],
            data="",
            ignore_errors=[404],
            error_counter=error_counter,
        )

        if status_code == 200:

            # Was deleted on local server, but found on target server,
            # which means we need to delete it
            if not data:
                yield _prepare_deleted_row(model, _id, error=True)
            # Was inserted or updated without errors
            elif checksum == _get_data_checksum(resp):
                conn.execute(
                    table.update().
                    where(
                        (table.c.id == _id)
                    ).
                    values(
                        revision=resp['_revision'],
                        error=False,
                        data=None
                    )
                )
                yield _PushRow(
                    model,
                    {'_type': type},
                    send=False
                )
                # Need to push again
            else:
                data['_revision'] = resp['_revision']
                yield _PushRow(
                    model,
                    data,
                    checksum=checksum,
                    saved=True,
                    op="patch",
                    error=True
                )

        elif status_code == 404:
            # Was deleted on both - local and target servers
            if not data:
                conn.execute(
                    table.delete().
                    where(
                        (table.c.id == _id)
                    )
                )
                yield _PushRow(
                    model,
                    {'_type': type},
                    send=False
                )
            # Need to push again
            else:
                yield _PushRow(
                    model,
                    data,
                    checksum=checksum,
                    saved=True,
                    op="insert",
                    error=True
                )


def _load_page_from_dict(model: Model, values: dict):
    model.page.clear()
    for key, item in values.items():
        prop = model.properties.get(key)
        if prop:
            model.page.add_prop(key, prop, item)



import re
from typing import List


def extract_error_property_names(error_message: str) -> List[str]:
    pattern = re.compile(r'Key \((.*?)\)')
    return pattern.search(error_message).group(1).split(".")[0].split(', ')


def extract_error_ref_id(error_message: str) -> str:
    pattern = re.compile(r'\)=\((.*?)\)')
    return pattern.search(error_message).group(1)


def extract_error_model(error_message: str) -> str:
    pattern = re.compile(r'\"([^"]*)\"')
    return pattern.search(error_message).group(1)



from spinta import commands
from spinta.components import Context
from spinta.datasets.backends.notimpl.components import BackendNotImplemented


@commands.wait.register(Context, BackendNotImplemented)
def wait(
    context: Context,
    backend: BackendNotImplemented,
    *,
    fail: bool = False,
) -> bool:
    return True



from sqlalchemy.dialects import registry
from sqlalchemy.dialects.sqlite.pysqlite import SQLiteDialect_pysqlite


class SQLiteDialect_spinta_sqlite(SQLiteDialect_pysqlite):
    driver = 'spinta_sqlite'
    supports_statement_cache = True

    @classmethod
    def dbapi(cls):
        try:
            from sqlean import dbapi2 as sqlite
        except ImportError:
            from sqlite3 import dbapi2 as sqlite
        return sqlite


registry.register(
    "sqlite.spinta_sqlite",
    __name__,
    "SQLiteDialect_spinta_sqlite"
)



from spinta import commands
from spinta.components import Context
from spinta.backends.nobackend.components import NoBackend


@commands.wait.register(Context, NoBackend)
def wait(context: Context, backend: NoBackend, *, fail: bool = False):
    return True



from typing import Any
from typing import Dict

from spinta import commands
from spinta.components import Context
from spinta.components import Node
from spinta.dimensions.prefix.components import UriPrefix
from spinta.nodes import load_node


@commands.load.register(Context, UriPrefix, dict)
def load(
    context: Context,
    prefix: UriPrefix,
    data: Dict[str, Any],
    *,
    parent: Node,
):
    return load_node(context, prefix, data, parent=parent)



from typing import AsyncIterator

from sqlalchemy import exc
from spinta import commands, exceptions
from spinta.commands import create_exception
from spinta.types.datatype import Denorm
from spinta.utils.data import take
from spinta.components import Context, Model, DataItem, DataSubItem
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.sqlalchemy import utcnow



@commands.insert.register(Context, Model, PostgreSQL)
async def insert(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    transaction = context.get('transaction')
    connection = transaction.connection
    table = backend.get_table(model)

    # Need to set specific amount of max errors, to prevent memory problems
    max_error_count = 100
    error_list = []
    savepoint_transaction_start = connection.begin_nested()
    rollback_full = False
    async for data in dstream:
        try:
            savepoint = connection.begin_nested()
            patch = commands.before_write(context, model, backend, data=data)
            # TODO: Refactor this to insert batches with single query.
            qry = table.insert().values(
                _id=patch['_id'],
                _revision=patch['_revision'],
                _txn=transaction.id,
                _created=utcnow(),
            )
            connection.execute(qry, patch)
            commands.after_write(context, model, backend, data=data)
        except exc.DatabaseError as error:
            rollback_full = True
            savepoint.rollback()
            exception = create_exception(data, error)
            error_list.append(exception)
            data.error = exception
            if len(error_list) >= max_error_count:
                yield data
                savepoint_transaction_start.rollback()
                raise exceptions.MultipleErrors(error_list)
        yield data

    if rollback_full:
        savepoint_transaction_start.rollback()
        raise exceptions.MultipleErrors(error_list)


@commands.update.register(Context, Model, PostgreSQL)
async def update(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    dstream: dict,
    stop_on_error: bool = True,
):
    transaction = context.get('transaction')
    connection = transaction.connection
    table = backend.get_table(model)

    # Need to set specific amount of max errors, to prevent memory problems
    max_error_count = 100
    error_list = []
    savepoint_transaction_start = connection.begin_nested()
    rollback_full = False
    async for data in dstream:
        if not data.patch:
            yield data
            continue
        try:
            savepoint = connection.begin_nested()
            pk = data.saved['_id']
            patch = commands.before_write(context, model, backend, data=data)
            result = connection.execute(
                table.update().
                where(table.c._id == pk).
                where(table.c._revision == data.saved['_revision']).
                values(patch)
            )
            if result.rowcount == 0:
                raise Exception(f"Update failed, {model} with {pk} not found.")
            elif result.rowcount > 1:
                raise Exception(
                    f"Update failed, {model} with {pk} has found and update "
                    f"{result.rowcount} rows."
                )
            commands.after_write(context, model, backend, data=data)
        except exc.DatabaseError as error:
            rollback_full = True
            savepoint.rollback()
            exception = create_exception(data, error)
            error_list.append(exception)
            data.error = exception

            if len(error_list) >= max_error_count:
                yield data
                savepoint_transaction_start.rollback()
                raise exceptions.MultipleErrors(error_list)
        yield data

    if rollback_full:
        savepoint_transaction_start.rollback()
        raise exceptions.MultipleErrors(error_list)


@commands.delete.register(Context, Model, PostgreSQL)
async def delete(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    dstream: AsyncIterator[DataItem],
    stop_on_error: bool = True,
):
    transaction = context.get('transaction')
    connection = transaction.connection
    table = backend.get_table(model)

    # Need to set specific amount of max errors, to prevent memory problems
    max_error_count = 100
    error_list = []
    savepoint_transaction_start = connection.begin_nested()
    rollback_full = False
    async for data in dstream:
        try:
            savepoint = connection.begin_nested()
            commands.before_write(context, model, backend, data=data)
            connection.execute(
                table.delete().
                where(table.c._id == data.saved['_id'])
            )
            commands.after_write(context, model, backend, data=data)
        except exc.DatabaseError as error:
            rollback_full = True
            savepoint.rollback()
            exception = create_exception(data, error)
            error_list.append(exception)
            data.error = exception

            if len(error_list) >= max_error_count:
                yield data
                savepoint_transaction_start.rollback()
                raise exceptions.MultipleErrors(error_list)
        yield data

    if rollback_full:
        savepoint_transaction_start.rollback()
        raise exceptions.MultipleErrors(error_list)


@commands.before_write.register(Context, Model, PostgreSQL)
def before_write(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    data: DataSubItem,
) -> dict:
    patch = take(['_id'], data.patch)
    patch['_revision'] = take('_revision', data.patch, data.saved)
    patch['_txn'] = context.get('transaction').id
    patch['_created'] = utcnow()
    for prop in take(model.properties).values():
        if isinstance(prop.dtype, Denorm):
            prop_data = data[prop.name.split('.')[0]]
        else:
            prop_data = data[prop.name]
        value = commands.before_write(
            context,
            prop.dtype,
            backend,
            data=prop_data,
        )
        patch.update(value)
    return patch


@commands.after_write.register(Context, Model, PostgreSQL)
def after_write(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    data: DataSubItem,
) -> dict:
    for key in take(data.patch or {}):
        prop = model.properties[key]
        commands.after_write(context, prop.dtype, backend, data=data[key])
